/* merdll.h - definitions for the merlin dll */

#ifndef __MERDLL_H__
#define __MERDLL_H__

#include "merlin.h"

#ifndef EXPORT
#ifdef WIN32
#include <windows.h>
#define EXPORT	__declspec(dllexport)
#else
#define EXPORT
#endif
#endif

#ifdef __MERLIN_INTERNAL__
#define structMMP	MMP
#define structMPE	MPE
#else
#define structMMP	MMPPublic
#define structMPE	MPEPublic
#endif

/* public mmp data */
typedef struct MMPPublic {
  long elapsedTicks;
  MScalar displayBase,displayTop;
  void (*displayHook)(struct structMMP *mmp,MScalar addr,MScalar value);
  void *displayData;
  int (*haltHook)(struct structMPE *mpe,MScalar flags,MScalar addr);
  void *haltData;
  void (*statHook)(struct structMMP *mmp,UINT32 *values);
  void *statData;
} MMPPublic;

/* public mpe data */
typedef struct MPEPublic {
  int dummy;
} MPEPublic;

#ifdef __MERLIN_INTERNAL__
#include "opcodes.h"
#include "memulate.h"
#else
typedef MMPPublic MMP;
typedef MPEPublic MPE;
#endif

/* prototypes */
EXPORT MMP *MakeMMP(int mpeCount,
                    int iRAMSize,int iROMSize,
                    int dRAMSize,int dROMSize,
                    int sRAMSize,int eRAMSize);
EXPORT void FreeMMP(MMP *mmp);
EXPORT MPE *mmpGetMPE(MMP *mmp,int n);
EXPORT int mmpMapAddress(MMP *mmp,MScalar addr,MScalar **pAddr);
EXPORT MMP *mpeGetMMP(MPE *mpe);
EXPORT int mpeFetchScalar(MPE *mpe,MScalar addr,MScalar *pValue);
EXPORT int mpeStoreScalar(MPE *mpe,MScalar addr,MScalar value);
EXPORT int mpeInitScalar(MPE *mpe,MScalar addr,MScalar value);
EXPORT int mpeLoadBinary(MPE *mpe,MScalar addr,char *fname);
EXPORT void mmpClock(MMP *mmp);
EXPORT int mpeRunState(MPE *mpe);
EXPORT int mpeSetRunState(MPE *mpe,int state);
EXPORT int mpeSingleStep(MPE *mpe);
EXPORT int mpeClock(MPE *mpe);
EXPORT int mpeRegisterValue(MPE *mpe,char *name,MScalar *pValue);
EXPORT int mpeSetRegisterValue(MPE *mpe,char *name,MScalar value);
EXPORT void mpeAccumulatorValue(MPE *mpe,MScalar *pValueHigh,MScalar *pValueLow);
EXPORT void mpeSetAccumulatorValue(MPE *mpe,MScalar valueHigh,MScalar valueLow);
EXPORT int mpeBreakpointP(MPE *mpe,MScalar addr);
EXPORT int mpeSetBreakpoint(MPE *mpe,MScalar addr);
EXPORT int mpeClearBreakpoint(MPE *mpe,MScalar addr);
EXPORT int mpeStoreInstruction(MPE *mpe,MScalar addr,UINT32 flags,char *text);
EXPORT void mul64(MScalar a,MScalar b,MScalar *pHigh,MScalar *pLow);
EXPORT void neg64(MScalar *pv1,MScalar *pv0);
EXPORT void add32(MScalar a,MScalar b,MScalar *pSum,int *pC,int *pV);
EXPORT void sub32(MScalar a,MScalar b,MScalar *pSum,int *pC,int *pV);
EXPORT void msb32(MScalar a,MScalar *pR);
EXPORT void bits32(MScalar n1,MScalar n0,int shift,MScalar *pR,int *pV);
EXPORT void shift64(MScalar n1,MScalar n0,int shift,MScalar *pHigh,MScalar *pLow);
EXPORT void ushift64(MScalar n1,MScalar n0,int shift,MScalar *pHigh,MScalar *pLow);
EXPORT void shift32(MScalar n,int shift,MScalar *pValue);
EXPORT void rotate32(MScalar n,int shift,MScalar *pValue);

#endif
