/* merlin.h - merlin definitions */
/*
	Copyright (c) 1995, by VM Labs, Inc.
	All Rights Reserved
*/

#ifndef __MERLIN_H__
#define __MERLIN_H__

/* data types */
typedef long MScalar;
typedef unsigned long UMScalar;
typedef unsigned long UINT32;

/* data type sizes */
#define bytesPerScalar	    4
#define bytesPerVector	    16
#define bytesPerInstruction 2

/* internal memory map */
#define sharedRAMBase       0x00200000
#define sharedRAMSize       (1024L * 1024L)
#define localRegBase	    0x20000000
#define localRegSize        256
#define localROMBase        0x20100000
#define localROMSize        (1024L * 1024L)
#define localRAMBase        0x20200000
#define localRAMSize        (1024L * 1024L)

/* instruction memory map */
#define instructionROMBase  0x20300000
#define instructionROMSize  (1024L * 1024L)
#define instructionRAMBase  0x20400000
#define instructionRAMSize  (1024L * 1024L)

/* external memory */
#define externalRAMBase     0x40000000
#define externalRAMSize     (2L * 1024L * 1024L)

/* instruction flags */
#define ifStop              0x80000000  /* end of compound instruction */
#define ifBreak             0x40000000  /* breakpoint, before/after */
#define if32Bits            0x20000000  /* 32 bit instruction */
#define if48Bits            0x10000000  /* 48 bit instruction */

/* condition codes */
#define ccT     0
#define ccEQ    1
#define ccNE    2
#define ccLT    3
#define ccLE    4
#define ccGT    5
#define ccGE    6
#define ccCC    7
#define ccCS    8
#define ccVC    9
#define ccVS    10
#define ccC0EQ  11
#define ccC0NE  12
#define ccC1NE  13
#define ccMVS   14
#define ccMVC   15

/* condition code bits */
#define ccZ     0x00000001
#define ccC     0x00000002
#define ccN     0x00000004
#define ccV     0x00000008
#define ccMV    0x00000010
#define ccC0Z   0x00000020
#define ccC1Z   0x00000040

/* stallops */
#define soNOP   0
#define soNEXT  1

/* run state values */
#define rsStop  0
#define rsRun   1

/* dma status bits */
#define dsPending   1
#define dsActive    2

/* halt hook codes */
#define hhHalt          0
#define hhBreak         1
#define hhError         2
#define hhBadLoad       3
#define hhBadMultiply   4

#endif
