/* opcodes.c - opcode handlers */

#include <stdio.h>
#define __MERLIN_INTERNAL__
#include "merdll.h"

/* local functions */
static void add(MPE *mpe,MScalar a,MScalar b,scalarRegister *r);
static void sub(MPE *mpe,MScalar a,MScalar b,scalarRegister *r);
static void cmp(MPE *mpe,MScalar a,MScalar b);
static void and(MPE *mpe,MScalar a,MScalar b,scalarRegister *r);
static void or(MPE *mpe,MScalar a,MScalar b,scalarRegister *r);
static void eor(MPE *mpe,MScalar a,MScalar b,scalarRegister *r);
static void asl(MPE *mpe,int shift,MScalar b,scalarRegister *r);
static void asr(MPE *mpe,int shift,MScalar b,scalarRegister *r);
static void lsr(MPE *mpe,int shift,MScalar b,scalarRegister *r);
static void msb(MPE *mpe,MScalar a,MScalar b,scalarRegister *r);
static void btst(MPE *mpe,MScalar a,MScalar b);
static void dotp_sr(MPE *mpe,MScalar a,scalarRegister *b,int shift,scalarRegister *r);
static void dotp_vr(MPE *mpe,scalarRegister *a,scalarRegister *b,int shift,scalarRegister *r);
static void mul(MPE *mpe,MScalar a,MScalar b,int shift,scalarRegister *r);
static void mul_p_sr(MPE *mpe,MScalar a,scalarRegister *b,int shift,scalarRegister *r);
static void mul_p_vr(MPE *mpe,scalarRegister *a,scalarRegister *b,int shift,scalarRegister *r);
static void mul_sv_sr(MPE *mpe,MScalar a,scalarRegister *b,int shift,scalarRegister *r);
static void mul_sv_vr(MPE *mpe,scalarRegister *a,scalarRegister *b,int shift,scalarRegister *r);
static int ShiftAmount(MScalar code);
static void dec(MPE *mpe,UINT32 *args);
static void ComputePixelAddress(bilinearRegister *r,int scale,MScalar *pAddr);
static void Branch(MPE *mpe,int cc,int type,MScalar target,int stallop);

/* nop */

static void h_nop(MPE *mpe,UINT32 *args)
{
    /* do nothing */
}

/* abs */
static void h_abs_sr(MPE *mpe,UINT32 *args)
{
    scalarRegister *r = (scalarRegister *)args[0];
    MScalar a = r->value;

    /* compute the absolute value */
    if (a < 0)
        r->newValue = -a;

    /* set the condition codes */
    mpe->cc.newValue &= ~(ccZ | ccC | ccN | ccV);
    if (a < 0)            mpe->cc.newValue |= ccC;
    if (r->newValue < 0)  mpe->cc.newValue |= ccN;
    if (r->newValue == 0) mpe->cc.newValue |= ccZ;
}

/* add sr,sr */
static void h_add_sr_sr(MPE *mpe,UINT32 *args)
{
    add(mpe,
        ((scalarRegister *)args[0])->value,
        ((scalarRegister *)args[1])->value,
        (scalarRegister *)args[1]);
}

/* add #n,sr */
static void h_add_uimm_sr(MPE *mpe,UINT32 *args)
{
    add(mpe,
        (MScalar)args[0],
        ((scalarRegister *)args[1])->value,
        (scalarRegister *)args[1]);
}

/* add sr,sr,sr */
static void h_add_sr_sr_sr(MPE *mpe,UINT32 *args)
{
    add(mpe,
        ((scalarRegister *)args[0])->value,
        ((scalarRegister *)args[1])->value,
        (scalarRegister *)args[2]);
}

/* add #n,sr,sr */
static void h_add_uimm_sr_sr(MPE *mpe,UINT32 *args)
{
    add(mpe,
        (MScalar)args[0],
        ((scalarRegister *)args[1])->value,
        (scalarRegister *)args[2]);
}

/* add #n,sr */
static void h_add_uimm10_sr(MPE *mpe,UINT32 *args)
{
    add(mpe,
        (MScalar)args[0],
        ((scalarRegister *)args[1])->value,
        (scalarRegister *)args[1]);
}

/* add #n,<>#n,sr */
static void h_add_uimm_rotate_imm_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    rotate32((MScalar)args[0],
             (int)args[1],
             &tmp);
    add(mpe,
        tmp,
        ((scalarRegister *)args[2])->value,
        (scalarRegister *)args[2]);
}

/* add sr,#n,sr */
static void h_add_sr_uimm_sr(MPE *mpe,UINT32 *args)
{
    add(mpe,
        ((scalarRegister *)args[0])->value,
        (MScalar)args[1],
        (scalarRegister *)args[2]);
}

/* add #n,>>sr,sr */
static void h_add_uimm_shift_sr_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    shift32((MScalar)args[0],
            (int)((scalarRegister *)args[1])->value,
            &tmp);
    add(mpe,
        tmp,
        ((scalarRegister *)args[2])->value,
        (scalarRegister *)args[2]);
}

/* add sr,>>#n,sr */
static void h_add_sr_shift_imm_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    shift32(((scalarRegister *)args[0])->value,
            (int)args[1],
            &tmp);
    add(mpe,
        tmp,
        ((scalarRegister *)args[2])->value,
        (scalarRegister *)args[2]);
}

/* add sr,>>sr,sr */
static void h_add_sr_shift_sr_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    shift32(((scalarRegister *)args[0])->value,
            (int)((scalarRegister *)args[1])->value,
            &tmp);
    add(mpe,
        tmp,
        ((scalarRegister *)args[2])->value,
        (scalarRegister *)args[2]);
}

/* add sr,<>sr,sr */
static void h_add_sr_rotate_sr_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    rotate32(((scalarRegister *)args[0])->value,
            (int)((scalarRegister *)args[1])->value,
            &tmp);
    add(mpe,
        tmp,
        ((scalarRegister *)args[2])->value,
        (scalarRegister *)args[2]);
}

/* add - common add routine */
static void add(MPE *mpe,MScalar a,MScalar b,scalarRegister *r)
{
    int c,v;

    /* do the add */
    add32(a,b,&r->newValue,&c,&v);

    /* set the condition codes */
    mpe->cc.newValue &= ~(ccZ | ccC | ccN | ccV);
    if (c)                mpe->cc.newValue |= ccC;
    if (v)                mpe->cc.newValue |= ccV;
    if (r->newValue < 0)  mpe->cc.newValue |= ccN;
    if (r->newValue == 0) mpe->cc.newValue |= ccZ;
}

/* add_sv vr,vr */
static void h_add_sv_vr_vr(MPE *mpe,UINT32 *args)
{
    scalarRegister *a = (scalarRegister *)args[0];
    scalarRegister *b = (scalarRegister *)args[1];
    scalarRegister *r = (scalarRegister *)args[1];
    int i = 4;
    while (--i >= 0)
        (*r++).newValue = ((*a++).value & 0xFFFF0000)
                        + ((*b++).value & 0xFFFF0000);
}

/* add_sv vr,vr,vr */
static void h_add_sv_vr_vr_vr(MPE *mpe,UINT32 *args)
{
    scalarRegister *a = (scalarRegister *)args[0];
    scalarRegister *b = (scalarRegister *)args[1];
    scalarRegister *r = (scalarRegister *)args[2];
    int i = 4;
    while (--i >= 0)
        (*r++).newValue = ((*a++).value & 0xFFFF0000)
                        + ((*b++).value & 0xFFFF0000);
}

/* sub sr,sr */
static void h_sub_sr_sr(MPE *mpe,UINT32 *args)
{
    sub(mpe,
        ((scalarRegister *)args[0])->value,
        ((scalarRegister *)args[1])->value,
        (scalarRegister *)args[1]);
}

/* sub #n,sr */
static void h_sub_uimm_sr(MPE *mpe,UINT32 *args)
{
    sub(mpe,
        (MScalar)args[0],
        ((scalarRegister *)args[1])->value,
        (scalarRegister *)args[1]);
}

/* sub sr,sr,sr */
static void h_sub_sr_sr_sr(MPE *mpe,UINT32 *args)
{
    sub(mpe,
        ((scalarRegister *)args[0])->value,
        ((scalarRegister *)args[1])->value,
        (scalarRegister *)args[2]);
}

/* sub #n,sr,sr */
static void h_sub_uimm_sr_sr(MPE *mpe,UINT32 *args)
{
    sub(mpe,
        (MScalar)args[0],
        ((scalarRegister *)args[1])->value,
        (scalarRegister *)args[2]);
}

/* sub #n,sr */
static void h_sub_uimm10_sr(MPE *mpe,UINT32 *args)
{
    sub(mpe,
        (MScalar)args[0],
        ((scalarRegister *)args[1])->value,
        (scalarRegister *)args[1]);
}

/* sub #n,<>#n,sr */
static void h_sub_uimm_rotate_imm_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    rotate32((MScalar)args[0],
             (int)args[1],
             &tmp);
    sub(mpe,
        tmp,
        ((scalarRegister *)args[2])->value,
        (scalarRegister *)args[2]);
}

/* sub sr,#n,sr */
static void h_sub_sr_uimm_sr(MPE *mpe,UINT32 *args)
{
    sub(mpe,
        ((scalarRegister *)args[0])->value,
        (MScalar)args[1],
        (scalarRegister *)args[2]);
}

/* sub #n,>>sr,sr */
static void h_sub_uimm_shift_sr_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    shift32((MScalar)args[0],
            (int)((scalarRegister *)args[1])->value,
            &tmp);
    sub(mpe,
        tmp,
        ((scalarRegister *)args[2])->value,
        (scalarRegister *)args[2]);
}

/* sub sr,>>#n,sr */
static void h_sub_sr_shift_imm_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    shift32(((scalarRegister *)args[0])->value,
            (int)args[1],
            &tmp);
    sub(mpe,
        tmp,
        ((scalarRegister *)args[2])->value,
        (scalarRegister *)args[2]);
}

/* sub sr,>>sr,sr */
static void h_sub_sr_shift_sr_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    shift32(((scalarRegister *)args[0])->value,
            (int)((scalarRegister *)args[1])->value,
            &tmp);
    sub(mpe,
        tmp,
        ((scalarRegister *)args[2])->value,
        (scalarRegister *)args[2]);
}

/* sub - common sub routine */
static void sub(MPE *mpe,MScalar a,MScalar b,scalarRegister *r)
{
    int c,v;

    /* do the subtract */
    sub32(a,b,&r->newValue,&c,&v);

    /* set the condition codes */
    mpe->cc.newValue &= ~(ccZ | ccC | ccN | ccV);
    if (c)                mpe->cc.newValue |= ccC;
    if (v)                mpe->cc.newValue |= ccV;
    if (r->newValue < 0)  mpe->cc.newValue |= ccN;
    if (r->newValue == 0) mpe->cc.newValue |= ccZ;
}

/* sub_sv vr,vr */
static void h_sub_sv_vr_vr(MPE *mpe,UINT32 *args)
{
    scalarRegister *a = (scalarRegister *)args[0];
    scalarRegister *b = (scalarRegister *)args[1];
    scalarRegister *r = (scalarRegister *)args[1];
    int i = 4;
    while (--i >= 0)
        (*r++).newValue = ((*b++).value & 0xFFFF0000)
                        - ((*a++).value & 0xFFFF0000);
}

/* sub_sv vr,vr,vr */
static void h_sub_sv_vr_vr_vr(MPE *mpe,UINT32 *args)
{
    scalarRegister *a = (scalarRegister *)args[0];
    scalarRegister *b = (scalarRegister *)args[1];
    scalarRegister *r = (scalarRegister *)args[2];
    int i = 4;
    while (--i >= 0)
        (*r++).newValue = ((*b++).value & 0xFFFF0000)
                        - ((*a++).value & 0xFFFF0000);
}

/* cmp sr,sr */
static void h_cmp_sr_sr(MPE *mpe,UINT32 *args)
{
    cmp(mpe,
        ((scalarRegister *)args[0])->value,
        ((scalarRegister *)args[1])->value);
}

/* cmp #n,sr */
static void h_cmp_uimm_sr(MPE *mpe,UINT32 *args)
{
    cmp(mpe,
        (MScalar)args[0],
        ((scalarRegister *)args[1])->value);
}

/* cmp #n,sr */
static void h_cmp_uimm10_sr(MPE *mpe,UINT32 *args)
{
    cmp(mpe,
        (MScalar)args[0],
        ((scalarRegister *)args[1])->value);
}

/* cmp #n,<>sr,sr */
static void h_cmp_uimm_rotate_imm_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    rotate32((MScalar)args[0],
             (int)args[1],
             &tmp);
    cmp(mpe,
        tmp,
        ((scalarRegister *)args[2])->value);
}

/* cmp sr,#n */
static void h_cmp_sr_uimm(MPE *mpe,UINT32 *args)
{
    cmp(mpe,
        ((scalarRegister *)args[0])->value,
        (MScalar)args[1]);
}

/* cmp #n,>>sr,sr */
static void h_cmp_uimm_shift_sr_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    shift32((MScalar)args[0],
            (int)((scalarRegister *)args[1])->value,
            &tmp);
    cmp(mpe,
        tmp,
        ((scalarRegister *)args[2])->value);
}

/* cmp #n,<>#n,sr */
static void h_cmp_imm_rotate_imm_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    rotate32((MScalar)args[0],
             (int)args[1],
             &tmp);
    cmp(mpe,
        tmp,
        ((scalarRegister *)args[2])->value);
}

/* cmp sr,>>#n,sr */
static void h_cmp_sr_shift_imm_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    shift32(((scalarRegister *)args[0])->value,
            (int)args[1],
            &tmp);
    cmp(mpe,
        tmp,
        ((scalarRegister *)args[2])->value);
}

/* cmp sr,>>sr,sr */
static void h_cmp_sr_shift_sr_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    shift32(((scalarRegister *)args[0])->value,
            (int)((scalarRegister *)args[1])->value,
            &tmp);
    cmp(mpe,
        tmp,
        ((scalarRegister *)args[2])->value);
}

/* cmp - common compare routine */
static void cmp(MPE *mpe,MScalar a,MScalar b)
{
    MScalar r;
    int c,v;

    /* do the subtract */
    sub32(a,b,&r,&c,&v);

    /* set the condition codes */
    mpe->cc.newValue &= ~(ccZ | ccC | ccN | ccV);
    if (c)      mpe->cc.newValue |= ccC;
    if (v)      mpe->cc.newValue |= ccV;
    if (r < 0)  mpe->cc.newValue |= ccN;
    if (r == 0) mpe->cc.newValue |= ccZ;
}

/* and sr,sr */
static void h_and_sr_sr(MPE *mpe,UINT32 *args)
{
    and(mpe,
        ((scalarRegister *)args[0])->value,
        ((scalarRegister *)args[1])->value,
        (scalarRegister *)args[1]);
}

/* and sr,sr,sr */
static void h_and_sr_sr_sr(MPE *mpe,UINT32 *args)
{
    and(mpe,
        ((scalarRegister *)args[0])->value,
        ((scalarRegister *)args[1])->value,
        (scalarRegister *)args[2]);
}

/* and #n,sr,sr */
static void h_and_imm_sr_sr(MPE *mpe,UINT32 *args)
{
    and(mpe,
        (MScalar)args[0],
        ((scalarRegister *)args[1])->value,
        (scalarRegister *)args[2]);
}

/* and #n,sr */
static void h_and_imm_sr(MPE *mpe,UINT32 *args)
{
    and(mpe,
        (MScalar)args[0],
        ((scalarRegister *)args[1])->value,
        (scalarRegister *)args[1]);
}

/* and #n,<>#n,sr */
static void h_and_imm_rotate_imm_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    rotate32((MScalar)args[0],
             (int)args[1],
             &tmp);
    and(mpe,
        tmp,
        ((scalarRegister *)args[2])->value,
        (scalarRegister *)args[2]);
}

/* and #n,>>sr,sr */
static void h_and_imm_shift_sr_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    shift32((MScalar)args[0],
            (int)((scalarRegister *)args[1])->value,
            &tmp);
    and(mpe,
        tmp,
        ((scalarRegister *)args[2])->value,
        (scalarRegister *)args[2]);
}

/* and #n,>>#n,sr */
static void h_and_imm_shift_imm_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    shift32((MScalar)args[0],
            (int)args[1],
            &tmp);
    and(mpe,
        tmp,
        ((scalarRegister *)args[2])->value,
        (scalarRegister *)args[2]);
}

/* and sr,>>#n,sr */
static void h_and_sr_shift_imm_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    shift32(((scalarRegister *)args[0])->value,
            (int)args[1],
            &tmp);
    and(mpe,
        tmp,
        ((scalarRegister *)args[2])->value,
        (scalarRegister *)args[2]);
}

/* and sr,>>sr,sr */
static void h_and_sr_shift_sr_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    shift32(((scalarRegister *)args[0])->value,
            (int)((scalarRegister *)args[1])->value,
            &tmp);
    and(mpe,
        tmp,
        ((scalarRegister *)args[2])->value,
        (scalarRegister *)args[2]);
}

/* and sr,<>sr,sr */
static void h_and_sr_rotate_sr_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    rotate32(((scalarRegister *)args[0])->value,
            (int)((scalarRegister *)args[1])->value,
            &tmp);
    and(mpe,
        tmp,
        ((scalarRegister *)args[2])->value,
        (scalarRegister *)args[2]);
}

/* and - common and routine */
static void and(MPE *mpe,MScalar a,MScalar b,scalarRegister *r)
{
    /* do the and */
    r->newValue = a & b;

    /* set the condition codes */
    mpe->cc.newValue &= ~(ccZ | ccN);
    if (r->newValue < 0)  mpe->cc.newValue |= ccN;
    if (r->newValue == 0) mpe->cc.newValue |= ccZ;
}

/* or sr,sr */
static void h_or_sr_sr(MPE *mpe,UINT32 *args)
{
    or(mpe,
       ((scalarRegister *)args[0])->value,
       ((scalarRegister *)args[1])->value,
       (scalarRegister *)args[1]);
}

/* or sr,sr,sr */
static void h_or_sr_sr_sr(MPE *mpe,UINT32 *args)
{
    or(mpe,
       ((scalarRegister *)args[0])->value,
       ((scalarRegister *)args[1])->value,
       (scalarRegister *)args[2]);
}

/* or #n,sr,sr */
static void h_or_imm_sr_sr(MPE *mpe,UINT32 *args)
{
    or(mpe,
       (MScalar)args[0],
       ((scalarRegister *)args[1])->value,
       (scalarRegister *)args[2]);
}

/* or #n,sr */
static void h_or_imm_sr(MPE *mpe,UINT32 *args)
{
    or(mpe,
       (MScalar)args[0],
       ((scalarRegister *)args[1])->value,
       (scalarRegister *)args[1]);
}

/* or #n,<>#n,sr */
static void h_or_imm_rotate_imm_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    rotate32((MScalar)args[0],
             (int)args[1],
             &tmp);
    or(mpe,
       tmp,
       ((scalarRegister *)args[2])->value,
       (scalarRegister *)args[2]);
}

/* or #n,>>sr,sr */
static void h_or_imm_shift_sr_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    shift32((MScalar)args[0],
            (int)((scalarRegister *)args[1])->value,
            &tmp);
    or(mpe,
       tmp,
       ((scalarRegister *)args[2])->value,
       (scalarRegister *)args[2]);
}

/* or #n,>>#n,sr */
static void h_or_imm_shift_imm_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    shift32((MScalar)args[0],
            (int)args[1],
            &tmp);
    or(mpe,
       tmp,
       ((scalarRegister *)args[2])->value,
       (scalarRegister *)args[2]);
}

/* or sr,>>#n,sr */
static void h_or_sr_shift_imm_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    shift32(((scalarRegister *)args[0])->value,
            (int)args[1],
            &tmp);
    or(mpe,
       tmp,
       ((scalarRegister *)args[2])->value,
       (scalarRegister *)args[2]);
}

/* or sr,>>sr,sr */
static void h_or_sr_shift_sr_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    shift32(((scalarRegister *)args[0])->value,
            (int)((scalarRegister *)args[1])->value,
            &tmp);
    or(mpe,
       tmp,
       ((scalarRegister *)args[2])->value,
       (scalarRegister *)args[2]);
}

/* or sr,<>sr,sr */
static void h_or_sr_rotate_sr_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    rotate32(((scalarRegister *)args[0])->value,
            (int)((scalarRegister *)args[1])->value,
            &tmp);
    or(mpe,
       tmp,
       ((scalarRegister *)args[2])->value,
       (scalarRegister *)args[2]);
}

/* or - common or routine */
static void or(MPE *mpe,MScalar a,MScalar b,scalarRegister *r)
{
    /* do the and */
    r->newValue = a | b;

    /* set the condition codes */
    mpe->cc.newValue &= ~(ccZ | ccN);
    if (r->newValue < 0)  mpe->cc.newValue |= ccN;
    if (r->newValue == 0) mpe->cc.newValue |= ccZ;
}

/* eor sr,sr */
static void h_eor_sr_sr(MPE *mpe,UINT32 *args)
{
    eor(mpe,
        ((scalarRegister *)args[0])->value,
        ((scalarRegister *)args[1])->value,
        (scalarRegister *)args[1]);
}

/* eor sr,sr,sr */
static void h_eor_sr_sr_sr(MPE *mpe,UINT32 *args)
{
    eor(mpe,
        ((scalarRegister *)args[0])->value,
        ((scalarRegister *)args[1])->value,
        (scalarRegister *)args[2]);
}

/* eor #n,sr,sr */
static void h_eor_imm_sr_sr(MPE *mpe,UINT32 *args)
{
    eor(mpe,
        (MScalar)args[0],
        ((scalarRegister *)args[1])->value,
        (scalarRegister *)args[2]);
}

/* eor #n,sr */
static void h_eor_imm_sr(MPE *mpe,UINT32 *args)
{
    eor(mpe,
        (MScalar)args[0],
        ((scalarRegister *)args[1])->value,
        (scalarRegister *)args[1]);
}

/* eor #n,<>#n,sr */
static void h_eor_imm_rotate_imm_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    rotate32((MScalar)args[0],
             (int)args[1],
             &tmp);
    eor(mpe,
        tmp,
        ((scalarRegister *)args[2])->value,
        (scalarRegister *)args[2]);
}

/* eor #n,>>sr,sr */
static void h_eor_imm_shift_sr_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    shift32((MScalar)args[0],
            (int)((scalarRegister *)args[1])->value,
            &tmp);
    eor(mpe,
        tmp,
        ((scalarRegister *)args[2])->value,
        (scalarRegister *)args[2]);
}

/* eor #n,>>#n,sr */
static void h_eor_imm_shift_imm_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    shift32((MScalar)args[0],
            (int)args[1],
            &tmp);
    eor(mpe,
        tmp,
        ((scalarRegister *)args[2])->value,
        (scalarRegister *)args[2]);
}

/* eor sr,>>#n,sr */
static void h_eor_sr_shift_imm_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    shift32(((scalarRegister *)args[0])->value,
            (int)args[1],
            &tmp);
    eor(mpe,
        tmp,
        ((scalarRegister *)args[2])->value,
        (scalarRegister *)args[2]);
}

/* eor sr,>>sr,sr */
static void h_eor_sr_shift_sr_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    shift32(((scalarRegister *)args[0])->value,
            (int)((scalarRegister *)args[1])->value,
            &tmp);
    eor(mpe,
        tmp,
        ((scalarRegister *)args[2])->value,
        (scalarRegister *)args[2]);
}

/* eor sr,<>sr,sr */
static void h_eor_sr_rotate_sr_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    rotate32(((scalarRegister *)args[0])->value,
            (int)((scalarRegister *)args[1])->value,
            &tmp);
    eor(mpe,
        tmp,
        ((scalarRegister *)args[2])->value,
        (scalarRegister *)args[2]);
}

/* eor - common eor routine */
static void eor(MPE *mpe,MScalar a,MScalar b,scalarRegister *r)
{
    /* do the and */
    r->newValue = a ^ b;

    /* set the condition codes */
    mpe->cc.newValue &= ~(ccZ | ccN);
    if (r->newValue < 0)  mpe->cc.newValue |= ccN;
    if (r->newValue == 0) mpe->cc.newValue |= ccZ;
}

/* neg sr */
static void h_neg_sr(MPE *mpe,UINT32 *args)
{
    sub(mpe,
        ((scalarRegister *)args[0])->value,
        0,
        (scalarRegister *)args[0]);
}

/* not sr */
static void h_not_sr(MPE *mpe,UINT32 *args)
{
    scalarRegister *r = (scalarRegister *)args[0];
    r->newValue = ~r->value;

    /* set the condition codes */
    mpe->cc.newValue &= ~(ccZ | ccN);
    if (r->newValue < 0)  mpe->cc.newValue |= ccN;
    if (r->newValue == 0) mpe->cc.newValue |= ccZ;
}

/* bits #n,>>#n,sr */
static void h_bits_uimm_shift_uimm_sr(MPE *mpe,UINT32 *args)
{
    MScalar mask = (1 << (MScalar)args[0]) - 1;
    scalarRegister *r = (scalarRegister *)args[2];
    r->newValue = (r->value >> (int)args[1]) & mask;
    mpe->cc.newValue &= ~(ccZ | ccN);
    if (r->newValue < 0)  mpe->cc.newValue |= ccN;
    if (r->newValue == 0) mpe->cc.newValue |= ccZ;
}

/* as sr,sr,sr */
static void h_as_sr_sr_sr(MPE *mpe,UINT32 *args)
{
    int shift = (int)((scalarRegister *)args[0])->value;
    scalarRegister *b = (scalarRegister *)args[1];
    scalarRegister *r = (scalarRegister *)args[2];
    asr(mpe,shift,b->value,r);
}

/* asl #n,sr */
static void h_asl_uimm_sr(MPE *mpe,UINT32 *args)
{
    int shift = (int)args[0];
    MScalar b = ((scalarRegister *)args[1])->value;
    scalarRegister *r = (scalarRegister *)args[1];
    asl(mpe,shift,b,r);
}

/* asl #n,sr,sr */
static void h_asl_uimm_sr_sr(MPE *mpe,UINT32 *args)
{
    int shift = (int)args[0];
    MScalar b = ((scalarRegister *)args[1])->value;
    scalarRegister *r = (scalarRegister *)args[2];
    asl(mpe,shift,b,r);
}

/* asl - common asl routine */
static void asl(MPE *mpe,int shift,MScalar b,scalarRegister *r)
{
    /* do the shift */
    r->newValue = shift < 0 ? b >> -shift : b << shift;
    
    /* set the condition codes */
    mpe->cc.newValue &= ~(ccC | ccZ | ccN);
    if (b & 0x80000000)   mpe->cc.newValue |= ccC;
    if (r->newValue < 0)  mpe->cc.newValue |= ccN;
    if (r->newValue == 0) mpe->cc.newValue |= ccZ;
}

/* asr #n,sr */
static void h_asr_uimm_sr(MPE *mpe,UINT32 *args)
{
    int shift = (int)args[0];
    MScalar b = ((scalarRegister *)args[1])->value;
    scalarRegister *r = (scalarRegister *)args[1];
    asr(mpe,shift,b,r);
}

/* asr #n,sr,sr */
static void h_asr_uimm_sr_sr(MPE *mpe,UINT32 *args)
{
    int shift = (int)args[0];
    MScalar b = ((scalarRegister *)args[1])->value;
    scalarRegister *r = (scalarRegister *)args[2];
    asr(mpe,shift,b,r);
}

/* asr - common asr routine */
static void asr(MPE *mpe,int shift,MScalar b,scalarRegister *r)
{
    /* do the shift */
    r->newValue = shift < 0 ? b << -shift : b >> shift;

    /* set the condition codes */
    mpe->cc.newValue &= ~(ccC | ccZ | ccN);
    if (b & 1)            mpe->cc.newValue |= ccC;
    if (r->newValue < 0)  mpe->cc.newValue |= ccN;
    if (r->newValue == 0) mpe->cc.newValue |= ccZ;
}

/* ls sr,sr,sr */
static void h_ls_sr_sr_sr(MPE *mpe,UINT32 *args)
{
    int shift = (int)((scalarRegister *)args[0])->value;
    scalarRegister *b = (scalarRegister *)args[1];
    scalarRegister *r = (scalarRegister *)args[2];
    lsr(mpe,shift,b->value,r);
}

/* lsl #n,sr */
static void h_lsl_uimm_sr(MPE *mpe,UINT32 *args)
{
    h_asl_uimm_sr(mpe,args);
}

/* lsl #n,sr,sr */
static void h_lsl_uimm_sr_sr(MPE *mpe,UINT32 *args)
{
    h_asl_uimm_sr_sr(mpe,args);
}

/* lsr #n,sr */
static void h_lsr_uimm_sr(MPE *mpe,UINT32 *args)
{
    int shift = (int)args[0];
    scalarRegister *b = (scalarRegister *)args[1];
    scalarRegister *r = (scalarRegister *)args[1];
    lsr(mpe,shift,b->value,r);
}

/* lsr #n,sr,sr */
static void h_lsr_uimm_sr_sr(MPE *mpe,UINT32 *args)
{
    int shift = (int)args[0];
    scalarRegister *b = (scalarRegister *)args[1];
    scalarRegister *r = (scalarRegister *)args[2];
    lsr(mpe,shift,b->value,r);
}

/* lsr - common lsr routine */
static void lsr(MPE *mpe,int shift,MScalar b,scalarRegister *r)
{
    /* do the shift */
    r->newValue = (MScalar)((UMScalar)b >> shift);

    /* set the condition codes */
    mpe->cc.newValue &= ~(ccC | ccZ | ccN);
    if (b & 1)            mpe->cc.newValue |= ccC;
    if (r->newValue < 0)  mpe->cc.newValue |= ccN;
    if (r->newValue == 0) mpe->cc.newValue |= ccZ;
}

/* msb sr,sr */
static void h_msb_sr_sr(MPE *mpe,UINT32 *args)
{
    msb(mpe,
        0,
        ((scalarRegister *)args[0])->value,
        (scalarRegister *)args[1]);
}

/* msb sr,sr,sr */
static void h_msb_sr_sr_sr(MPE *mpe,UINT32 *args)
{
    msb(mpe,
        ((scalarRegister *)args[0])->value,
        ((scalarRegister *)args[1])->value,
        (scalarRegister *)args[2]);
}

/* msb #n,sr,sr */
static void h_msb_imm_sr_sr(MPE *mpe,UINT32 *args)
{
    msb(mpe,
        (MScalar)args[0],
        ((scalarRegister *)args[1])->value,
        (scalarRegister *)args[2]);
}

/* msb - common msb routine */
static void msb(MPE *mpe,MScalar a,MScalar b,scalarRegister *r)
{
    int c,v;

    /* do the add */
    msb32(b,&b);
    add32(a,b,&r->newValue,&c,&v);

    /* set the condition codes */
    mpe->cc.newValue &= ~(ccZ | ccC | ccN | ccV);
    if (c)                mpe->cc.newValue |= ccC;
    if (v)                mpe->cc.newValue |= ccV;
    if (r->newValue < 0)  mpe->cc.newValue |= ccN;
    if (r->newValue == 0) mpe->cc.newValue |= ccZ;
}

/* bclr #n,sr */
static void h_bclr_uimm_sr(MPE *mpe,UINT32 *args)
{
    MScalar a = 1 << (int)args[0];
    scalarRegister *b = (scalarRegister *)args[1];
    MScalar r = b->value & ~a;

    /* store the new value */
    b->newValue = r;
    
    /* set the condition codes */
    mpe->cc.newValue &= ~(ccZ | ccN);
    if (r < 0)  mpe->cc.newValue |= ccN;
    if (r == 0) mpe->cc.newValue |= ccZ;
}

/* bset #n,sr */
static void h_bset_uimm_sr(MPE *mpe,UINT32 *args)
{
    MScalar a = 1 << (int)args[0];
    scalarRegister *b = (scalarRegister *)args[1];
    MScalar r = b->value | a;

    /* store the new value */
    b->newValue = r;
    
    /* set the condition codes */
    mpe->cc.newValue &= ~(ccZ | ccN);
    if (r < 0)  mpe->cc.newValue |= ccN;
    if (r == 0) mpe->cc.newValue |= ccZ;
}

/* btst #n,sr */
static void h_btst_uimm_sr(MPE *mpe,UINT32 *args)
{
    MScalar a = 1 << (int)args[0];
    MScalar b = ((scalarRegister *)args[1])->value;
    btst(mpe,a,b);
}

/* btst #n,<>,#n,sr */
static void h_btst_imm_rotate_imm_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    rotate32((MScalar)args[0],
             (int)args[1],
             &tmp);
    btst(mpe,
         tmp,
         ((scalarRegister *)args[2])->value);
}

/* btst #n,>>sr,sr */
static void h_btst_imm_shift_sr_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    shift32((MScalar)args[0],
            (int)((scalarRegister *)args[1])->value,
            &tmp);
    btst(mpe,
         tmp,
         ((scalarRegister *)args[2])->value);
}

/* btst sr,>>#n,sr */
static void h_btst_sr_shift_imm_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    shift32(((scalarRegister *)args[0])->value,
            (int)args[1],
            &tmp);
    btst(mpe,
         tmp,
         ((scalarRegister *)args[2])->value);
}

/* btst sr,>>sr,sr */
static void h_btst_sr_shift_sr_sr(MPE *mpe,UINT32 *args)
{
    MScalar tmp;
    shift32(((scalarRegister *)args[0])->value,
            (int)((scalarRegister *)args[1])->value,
            &tmp);
    btst(mpe,
         tmp,
         ((scalarRegister *)args[2])->value);
}

/* btst - common btst routine */
static void btst(MPE *mpe,MScalar a,MScalar b)
{
    MScalar r;
    
    /* do the and */
    r = a & b;

    /* set the condition codes */
    mpe->cc.newValue &= ~(ccZ | ccN);
    if (r < 0)  mpe->cc.newValue |= ccN;
    if (r == 0) mpe->cc.newValue |= ccZ;
}

/* copy sr,sr */
static void h_copy_sr_sr(MPE *mpe,UINT32 *args)
{
    scalarRegister *a = (scalarRegister *)args[0];
    scalarRegister *b = (scalarRegister *)args[1];
    b->newValue = a->value;
}

/* butt */
static void h_butt_sr_sr(MPE *mpe,UINT32 *args)
{
    scalarRegister *a = (scalarRegister *)args[0];
    scalarRegister *b = (scalarRegister *)args[1];
    add(mpe,a[0].value,a[1].value,&b[0]);
    b[1].newValue = a[1].value - a[0].value; 
}

/* butt */
static void h_butt_sr_sr_sr(MPE *mpe,UINT32 *args)
{
    scalarRegister *a = (scalarRegister *)args[0];
    scalarRegister *b = (scalarRegister *)args[1];
    scalarRegister *r = (scalarRegister *)args[2];
    add(mpe,a->value,b->value,&r[0]);
    r[1].newValue = b->value - a->value; 
}

/* dotp */
static void h_dotp_sr_vr_shift_sv_shift_sr(MPE *mpe,UINT32 *args)
{
    MScalar a = ((scalarRegister *)args[0])->value >> 16;
    scalarRegister *b = (scalarRegister *)args[1];
    int shift = ShiftAmount(((ioRegister *)args[2])->value);
    scalarRegister *r = (scalarRegister *)args[3];
    dotp_sr(mpe,a,b,shift,r);
}

/* dotp */
static void h_dotp_sr_vr_shift_mulimm_sr(MPE *mpe,UINT32 *args)
{
    MScalar a = ((scalarRegister *)args[0])->value >> 16;
    scalarRegister *b = (scalarRegister *)args[1];
    int shift = (MScalar)args[2];
    scalarRegister *r = (scalarRegister *)args[3];
    dotp_sr(mpe,a,b,shift,r);
}

/* dotp_sr - common dotp_sr routine */
static void dotp_sr(MPE *mpe,MScalar a,scalarRegister *b,int shift,scalarRegister *r)
{
    MScalar v;
    v  = (a * ((*b++).value >> 16)) << (16 - shift);
    v += (a * ((*b++).value >> 16)) << (16 - shift);
    v += (a * ((*b++).value >> 16)) << (16 - shift);
    v += (a * ((*b++).value >> 16)) << (16 - shift);
    StageScalarLoad(&mpe->multiplyBuffer,r,v);
}

/* dotp */
static void h_dotp_vr_vr_shift_sv_shift_sr(MPE *mpe,UINT32 *args)
{
    scalarRegister *a = (scalarRegister *)args[0];
    scalarRegister *b = (scalarRegister *)args[1];
    int shift = ShiftAmount(((ioRegister *)args[2])->value);
    scalarRegister *r = (scalarRegister *)args[3];
    dotp_vr(mpe,a,b,shift,r);
}

/* dotp */
static void h_dotp_vr_vr_shift_mulimm_sr(MPE *mpe,UINT32 *args)
{
    scalarRegister *a = (scalarRegister *)args[0];
    scalarRegister *b = (scalarRegister *)args[1];
    int shift = (MScalar)args[2];
    scalarRegister *r = (scalarRegister *)args[3];
    dotp_vr(mpe,a,b,shift,r);
}

/* dotp_vr - common dotp_vr routine */
static void dotp_vr(MPE *mpe,scalarRegister *a,scalarRegister *b,int shift,scalarRegister *r)
{
    MScalar v;
    v  = (((*a++).value >> 16)
       *  ((*b++).value >> 16)) << (16 - shift);
    v += (((*a++).value >> 16)
       *  ((*b++).value >> 16)) << (16 - shift);
    v += (((*a++).value >> 16)
       *  ((*b++).value >> 16)) << (16 - shift);
    v += (((*a++).value >> 16)
       *  ((*b++).value >> 16)) << (16 - shift);
    StageScalarLoad(&mpe->multiplyBuffer,r,v);
}

/* mul */
static void h_mul_sr_sr(MPE *mpe,UINT32 *args)
{
    mul(mpe,
        ((scalarRegister *)args[0])->value,
        ((scalarRegister *)args[1])->value,
        (int)mpe->ac_shift.value,
        ((scalarRegister *)args[1]));
}

/* mul */
static void h_mul_sr_sr_shift_ac_shift_sr(MPE *mpe,UINT32 *args)
{
    mul(mpe,
        ((scalarRegister *)args[0])->value,
        ((scalarRegister *)args[1])->value,
        ((ioRegister *)args[2])->value,
        ((scalarRegister *)args[3]));
}

/* mul */
static void h_mul_sr_sr_shift_uimm_sr(MPE *mpe,UINT32 *args)
{
    mul(mpe,
        ((scalarRegister *)args[0])->value,
        ((scalarRegister *)args[1])->value,
        (MScalar)args[2],
        ((scalarRegister *)args[3]));
}

/* mul */
static void h_mul_sr_sr_shift_sr_sr(MPE *mpe,UINT32 *args)
{
    mul(mpe,
        ((scalarRegister *)args[0])->value,
        ((scalarRegister *)args[1])->value,
        ((scalarRegister *)args[2])->value,
        ((scalarRegister *)args[3]));
}

/* mul */
static void h_mul_uimm_sr_shift_ac_shift_sr(MPE *mpe,UINT32 *args)
{
    mul(mpe,
        (MScalar)args[0],
        ((scalarRegister *)args[1])->value,
        ((ioRegister *)args[2])->value,
        ((scalarRegister *)args[3]));
}

/* mul */
static void h_mul_uimm_sr_shift_uimm_sr(MPE *mpe,UINT32 *args)
{
    mul(mpe,
        (MScalar)args[0],
        ((scalarRegister *)args[1])->value,
        (MScalar)args[2],
        ((scalarRegister *)args[3]));
}

/* mul */
static void h_mul_uimm_sr_shift_sr_sr(MPE *mpe,UINT32 *args)
{
    mul(mpe,
        (MScalar)args[0],
        ((scalarRegister *)args[1])->value,
        ((scalarRegister *)args[2])->value,
        ((scalarRegister *)args[3]));
}

/* mul - common mul routine */
static void mul(MPE *mpe,MScalar a,MScalar b,int shift,scalarRegister *r)
{
    MScalar high,low,v;
    int overflow;
    mul64(a,b,&high,&low);
    bits32(high,low,shift,&v,&overflow);
    mpe->ac.newValueHigh = high;
    mpe->ac.newValueLow = low;
    StageScalarLoad(&mpe->multiplyBuffer,r,v);
    mpe->multiplyOverflowBuffer = overflow ? mvOverflow : mvNoOverflow;
}

/* mul_p */
static void h_mul_p_sr_vr_shift_sv_shift_vr(MPE *mpe,UINT32 *args)
{
    MScalar a = ((scalarRegister *)args[0])->value >> 16;
    scalarRegister *b = (scalarRegister *)args[1];
    int shift = ShiftAmount(((ioRegister *)args[2])->value);
    scalarRegister *r = (scalarRegister *)args[3];
    mul_p_sr(mpe,a,b,shift,r);
}

/* mul_p */
static void h_mul_p_sr_vr_shift_mulimm_vr(MPE *mpe,UINT32 *args)
{
    MScalar a = ((scalarRegister *)args[0])->value >> 16;
    scalarRegister *b = (scalarRegister *)args[1];
    int shift = (int)args[2];
    scalarRegister *r = (scalarRegister *)args[3];
    mul_p_sr(mpe,a,b,shift,r);
}

/* mul_p_sr - common mul_p_sr routine */
static void mul_p_sr(MPE *mpe,MScalar a,scalarRegister *b,int shift,scalarRegister *r)
{
    MScalar v[3];
    v[0] = (a * ((*b++).value >> 16)) << (16 - shift);
    v[1] = (a * ((*b++).value >> 16)) << (16 - shift);
    v[2] = (a * ((*b++).value >> 16)) << (16 - shift);
    StagePixelLoad(&mpe->multiplyBuffer,r,v);
}

/* mul_p */
static void h_mul_p_ruv_vr_shift_sv_shift_vr(MPE *mpe,UINT32 *args)
{
    MScalar a = (((ioRegister *)args[0])->value & 0xfffc) >> 2;
    scalarRegister *b = (scalarRegister *)args[1];
    int shift = ShiftAmount(((ioRegister *)args[2])->value);
    scalarRegister *r = (scalarRegister *)args[3];
    mul_p_sr(mpe,a,b,shift,r);
}

/* mul_p */
static void h_mul_p_ruv_vr_shift_mulimm_vr(MPE *mpe,UINT32 *args)
{
    MScalar a = (((ioRegister *)args[0])->value & 0xfffc) >> 2;
    scalarRegister *b = (scalarRegister *)args[1];
    int shift = (int)args[2];
    scalarRegister *r = (scalarRegister *)args[3];
    mul_p_sr(mpe,a,b,shift,r);
}

/* mul_p */
static void h_mul_p_vr_vr_shift_sv_shift_vr(MPE *mpe,UINT32 *args)
{
    scalarRegister *a = (scalarRegister *)args[0];
    scalarRegister *b = (scalarRegister *)args[1];
    int shift = ShiftAmount(((ioRegister *)args[2])->value);
    scalarRegister *r = (scalarRegister *)args[3];
    mul_p_vr(mpe,a,b,shift,r);
}

/* mul_p */
static void h_mul_p_vr_vr_shift_mulimm_vr(MPE *mpe,UINT32 *args)
{
    scalarRegister *a = (scalarRegister *)args[0];
    scalarRegister *b = (scalarRegister *)args[1];
    int shift = (MScalar)args[2];
    scalarRegister *r = (scalarRegister *)args[3];
    mul_p_vr(mpe,a,b,shift,r);
}

/* mul_p_vr - common mul_p_vr routine */
static void mul_p_vr(MPE *mpe,scalarRegister *a,scalarRegister *b,int shift,scalarRegister *r)
{
    MScalar v[3];
    v[0] = (((*a++).value >> 16) * ((*b++).value >> 16)) << (16 - shift);
    v[1] = (((*a++).value >> 16) * ((*b++).value >> 16)) << (16 - shift);
    v[2] = (((*a++).value >> 16) * ((*b++).value >> 16)) << (16 - shift);
    StagePixelLoad(&mpe->multiplyBuffer,r,v);
}

/* mul_sv */
static void h_mul_sv_vr_vr(MPE *mpe,UINT32 *args)
{
    scalarRegister *a = (scalarRegister *)args[0];
    scalarRegister *b = (scalarRegister *)args[1];
    int shift = ShiftAmount(mpe->sv_shift.value);
    mul_sv_vr(mpe,a,b,shift,b);
}

/* mul_sv */
static void h_mul_sv_sr_vr_shift_sv_shift_vr(MPE *mpe,UINT32 *args)
{
    MScalar a = ((scalarRegister *)args[0])->value >> 16;
    scalarRegister *b = (scalarRegister *)args[1];
    int shift = ShiftAmount(((ioRegister *)args[2])->value);
    scalarRegister *r = (scalarRegister *)args[3];
    mul_sv_sr(mpe,a,b,shift,r);
}

/* mul_sv */
static void h_mul_sv_sr_vr_shift_mulimm_vr(MPE *mpe,UINT32 *args)
{
    MScalar a = ((scalarRegister *)args[0])->value >> 16;
    scalarRegister *b = (scalarRegister *)args[1];
    int shift = (MScalar)args[2];
    scalarRegister *r = (scalarRegister *)args[3];
    mul_sv_sr(mpe,a,b,shift,r);
}

/* mul_sv_sr - common mul_sv_sr routine */
static void mul_sv_sr(MPE *mpe,MScalar a,scalarRegister *b,int shift,scalarRegister *r)
{
    MScalar v[4];
    v[0] = (a * ((*b++).value >> 16)) << (16 - shift);
    v[1] = (a * ((*b++).value >> 16)) << (16 - shift);
    v[2] = (a * ((*b++).value >> 16)) << (16 - shift);
    v[3] = (a * ((*b++).value >> 16)) << (16 - shift);
    StageVectorLoad(&mpe->multiplyBuffer,r,v);
}

/* mul_sv */
static void h_mul_sv_ruv_vr_shift_sv_shift_vr(MPE *mpe,UINT32 *args)
{
    MScalar a = (((ioRegister *)args[0])->value & 0xfffc) >> 2;
    scalarRegister *b = (scalarRegister *)args[1];
    int shift = ShiftAmount(((ioRegister *)args[2])->value);
    scalarRegister *r = (scalarRegister *)args[3];
    mul_sv_sr(mpe,a,b,shift,r);
}

/* mul_sv */
static void h_mul_sv_ruv_vr_shift_mulimm_vr(MPE *mpe,UINT32 *args)
{
    MScalar a = (((ioRegister *)args[0])->value & 0xfffc) >> 2;
    scalarRegister *b = (scalarRegister *)args[1];
    int shift = (MScalar)args[2];
    scalarRegister *r = (scalarRegister *)args[3];
    mul_sv_sr(mpe,a,b,shift,r);
}

/* mul_sv */
static void h_mul_sv_vr_vr_shift_sv_shift_vr(MPE *mpe,UINT32 *args)
{
    scalarRegister *a = (scalarRegister *)args[0];
    scalarRegister *b = (scalarRegister *)args[1];
    int shift = ShiftAmount(((ioRegister *)args[2])->value);
    scalarRegister *r = (scalarRegister *)args[3];
    mul_sv_vr(mpe,a,b,shift,r);
}

/* mul_sv */
static void h_mul_sv_vr_vr_shift_mulimm_vr(MPE *mpe,UINT32 *args)
{
    scalarRegister *a = (scalarRegister *)args[0];
    scalarRegister *b = (scalarRegister *)args[1];
    int shift = (MScalar)args[2];
    scalarRegister *r = (scalarRegister *)args[3];
    mul_sv_vr(mpe,a,b,shift,r);
}

/* mul_sv_vr - common mul_sv_vr routine */
static void mul_sv_vr(MPE *mpe,scalarRegister *a,scalarRegister *b,int shift,scalarRegister *r)
{
    MScalar v[4];
    v[0] = (((*a++).value >> 16) * ((*b++).value >> 16)) << (16 - shift);
    v[1] = (((*a++).value >> 16) * ((*b++).value >> 16)) << (16 - shift);
    v[2] = (((*a++).value >> 16) * ((*b++).value >> 16)) << (16 - shift);
    v[3] = (((*a++).value >> 16) * ((*b++).value >> 16)) << (16 - shift);
    StageVectorLoad(&mpe->multiplyBuffer,r,v);
}

/* ShiftAmount - get the shift amount for a small vector multiply */
static int ShiftAmount(MScalar code)
{
    switch (code & 3) {
    case 0: return 0;
    case 1: return 8;
    case 2: return 12;
    case 3: return 14;
    }
}

/* addm */
static void h_addm_sr_sr_sr(MPE *mpe,UINT32 *args)
{
    MScalar a = ((scalarRegister *)args[0])->value;
    MScalar b = ((scalarRegister *)args[1])->value;
    scalarRegister *r = (scalarRegister *)args[2];
    r->newValue = a + b;
}

/* subm */
static void h_subm_sr_sr_sr(MPE *mpe,UINT32 *args)
{
    MScalar a = ((scalarRegister *)args[0])->value;
    MScalar b = ((scalarRegister *)args[1])->value;
    scalarRegister *r = (scalarRegister *)args[2];
    r->newValue = b - a;
}

/* buttm */
static void h_buttm_sr_sr(MPE *mpe,UINT32 *args)
{
    scalarRegister *a = (scalarRegister *)args[0];
    scalarRegister *b = (scalarRegister *)args[1];
    b[0].newValue = a[1].value + a[0].value; 
    b[1].newValue = a[1].value - a[0].value; 
}

/* buttm */
static void h_buttm_sr_sr_sr(MPE *mpe,UINT32 *args)
{
    scalarRegister *a = (scalarRegister *)args[0];
    scalarRegister *b = (scalarRegister *)args[1];
    scalarRegister *r = (scalarRegister *)args[2];
    r[0].newValue = b->value + a->value; 
    r[1].newValue = b->value - a->value; 
}

/* addr #n,rur,--rc0,--rc1 */
static void h_addr_ruimm_rur_rc_rc(MPE *mpe,UINT32 *args)
{
    ioRegister *b = (ioRegister *)args[1];
    b->newValue = b->value + args[0];
    dec(mpe,args);
}

/* addr sr,rur,--rc0,--rc1 */
static void h_addr_sr_rur_rc_rc(MPE *mpe,UINT32 *args)
{
    ioRegister *b = (ioRegister *)args[1];
    b->newValue = b->value + ((scalarRegister *)args[0])->value;
    dec(mpe,args);
}

/* mv_spr pc,rz,--rc0,--rc1 */
static void h_mv_spr_pc_rz_rc_rc(MPE *mpe,UINT32 *args)
{
    ioRegister *a = (ioRegister *)args[0];
    ioRegister *b = (ioRegister *)args[1];
    mpe->saveReturnP = TRUE;
    dec(mpe,args);
}

/* dec - common dec routine */
static void dec(MPE *mpe,UINT32 *args)
{
    ioRegister *rc0 = (ioRegister *)args[2];
    ioRegister *rc1 = (ioRegister *)args[3];
    if (rc0) {
        rc0->newValue = rc0->value - 1;
        mpe->cc.newValue &= ~ccC0Z;
        if (rc0->newValue == 0) mpe->cc.newValue |= ccC0Z;
    }
    if (rc1) {
        rc1->newValue = rc1->value - 1;
        mpe->cc.newValue &= ~ccC1Z;
        if (rc1->newValue == 0) mpe->cc.newValue |= ccC1Z;
    }
}

/* ld_s (sr),sr */
static void h_ld_s_ind_sr_sr(MPE *mpe,UINT32 *args)
{
    MScalar a = ((scalarRegister *)args[0])->value;
    scalarRegister *b = (scalarRegister *)args[1];
    MScalar value;
    mpeFetchScalar(mpe,a,&value);
    StageScalarLoad(&mpe->loadBuffer,b,value);
}

/* ld_s dt,sr */
static void h_ld_s_dsoff4_sr(MPE *mpe,UINT32 *args)
{
    MScalar a = (MScalar)args[1];
    scalarRegister *b = (scalarRegister *)args[2];
    MScalar value;
    a += (args[0] == dtRAMSpace ? localRAMBase : localROMBase);
    mpeFetchScalar(mpe,a,&value);
    StageScalarLoad(&mpe->loadBuffer,b,value);
}

/* ld_s (xy),sr */
static void h_ld_s_ind_bilinear_sr(MPE *mpe,UINT32 *args)
{
    CallHaltHook(mpe,hhError,mpe->pc.value);
}

/* ld_v (sr),vr */
static void h_ld_v_ind_sr_vr(MPE *mpe,UINT32 *args)
{
    MScalar a = ((scalarRegister *)args[0])->value;
    scalarRegister *b = (scalarRegister *)args[1];
    MScalar v[4];
    a &= vectorAddressMask;
    mpeFetchScalar(mpe,a,&v[0]);
    mpeFetchScalar(mpe,a + 4,&v[1]);
    mpeFetchScalar(mpe,a + 8,&v[2]);
    mpeFetchScalar(mpe,a + 12,&v[3]);
    StageVectorLoad(&mpe->loadBuffer,b,v);
}

/* ld_v dt,vr */
static void h_ld_v_dsoff16_vr(MPE *mpe,UINT32 *args)
{
    MScalar a = (MScalar)args[1];
    scalarRegister *b = (scalarRegister *)args[2];
    MScalar v[4];
    a += (args[0] == dtRAMSpace ? localRAMBase : localROMBase);
    a &= vectorAddressMask;
    mpeFetchScalar(mpe,a,&v[0]);
    mpeFetchScalar(mpe,a + 4,&v[1]);
    mpeFetchScalar(mpe,a + 8,&v[2]);
    mpeFetchScalar(mpe,a + 12,&v[3]);
    StageVectorLoad(&mpe->loadBuffer,b,v);
}

/* ld_sv (sr),vr */
static void h_ld_sv_ind_sr_vr(MPE *mpe,UINT32 *args)
{
    MScalar a = ((scalarRegister *)args[0])->value;
    scalarRegister *b = (scalarRegister *)args[1];
    MScalar v[4],value;
    a &= smallVectorAddressMask;
    mpeFetchScalar(mpe,a,&value);
    v[0] = value << 16;
    v[1] = value & 0xFFFF0000;
    mpeFetchScalar(mpe,a + 4,&value);
    v[2] = value << 16;
    v[3] = value & 0xFFFF0000;
    StageVectorLoad(&mpe->loadBuffer,b,v);
}

/* ld_sv dt,vr */
static void h_ld_sv_dsoff8_vr(MPE *mpe,UINT32 *args)
{
    MScalar a = (MScalar)args[1];
    scalarRegister *b = (scalarRegister *)args[2];
    MScalar v[4],value;
    a += (args[0] == dtRAMSpace ? localRAMBase : localROMBase);
    a &= smallVectorAddressMask;
    mpeFetchScalar(mpe,a,&value);
    v[0] = value << 16;
    v[1] = value & 0xFFFF0000;
    mpeFetchScalar(mpe,a + 4,&value);
    v[2] = value << 16;
    v[3] = value & 0xFFFF0000;
    StageVectorLoad(&mpe->loadBuffer,b,v);
}

/* ld_sv (xy),vr */
static void h_ld_sv_ind_bilinear_vr(MPE *mpe,UINT32 *args)
{
    CallHaltHook(mpe,hhError,mpe->pc.value);
}

/* ld_p (xy),pr */
static void h_ld_p_ind_bilinear_pr(MPE *mpe,UINT32 *args)
{
    scalarRegister *r = (scalarRegister *)args[1];
    MScalar ea,pixel,v[4];

    /* compute the pixel address and get the pixel */
    ComputePixelAddress((bilinearRegister *)args[0],4,&ea);
    mpeFetchScalar(mpe,ea,&pixel);

    /* decompose the pixel */
    v[0] = (pixel << 22) & 0x3fc00000;
    v[1] = (pixel << 14) & 0x3fc00000;
    v[2] = (pixel <<  6) & 0x3fc00000;
    v[3] = (pixel >>  2) & 0x3fc00000;

    /* normalize the chrominance values */
    if (TRUE) {
        v[1] -= 0x20000000;
        v[2] -= 0x20000000;
    }
    
    /* load the pixel into the vector register */
    StageVectorLoad(&mpe->loadBuffer,r,v);
}
/* ComputePixelAddress - compute the address of a pixel */
static void ComputePixelAddress(bilinearRegister *r,int scale,MScalar *pAddr)
{
    MScalar rowWidth = r->width->value & 0x7ff;
    MScalar xOffset = r->x->value >> 16;
    MScalar yOffset = r->y->value >> 16;
    MScalar mask = 0xffffffff >> (r->tile->value + 16);
    MScalar offset = (yOffset & mask) * rowWidth + (xOffset & mask);
    *pAddr = r->base->value + offset * scale;
}

/* ld_io (io),sr */
static void h_ld_io_ind_io_sr(MPE *mpe,UINT32 *args)
{
    ioRegister *a = (ioRegister *)args[0];
    scalarRegister *b = (scalarRegister *)args[1];
    if (a->readHandler)
        (*a->readHandler)(mpe);
    b->newValue = a->value;
}

/* ld_io (io),vr */
static void h_ld_io_ind_io_vr(MPE *mpe,UINT32 *args)
{
    ioVectorRegister *a = (ioVectorRegister *)args[0];
    scalarRegister *b = (scalarRegister *)args[1];
    ioRegister **p = a->data;
    if (a->readHandler)
        (*a->readHandler)(mpe);
    (*b++).newValue = (*p++)->value;
    (*b++).newValue = (*p++)->value;
    (*b++).newValue = (*p++)->value;
    (*b++).newValue = (*p++)->value;
}

/* st_s sr,(sr) */
static void h_st_s_sr_ind_sr(MPE *mpe,UINT32 *args)
{
    scalarRegister *a = (scalarRegister *)args[0];
    MScalar b = ((scalarRegister *)args[1])->value;
    mpeStoreScalar(mpe,b,a->newValue);
}

/* st_s sr,dt */
static void h_st_s_sr_dsoff4(MPE *mpe,UINT32 *args)
{
    scalarRegister *a = (scalarRegister *)args[0];
    MScalar b = (MScalar)args[2];
    b += (args[1] == dtRAMSpace ? localRAMBase : localROMBase);
    mpeStoreScalar(mpe,b,a->newValue);
}

/* st_s sr,(xy) */
static void h_st_s_sr_ind_bilinear(MPE *mpe,UINT32 *args)
{
    CallHaltHook(mpe,hhError,mpe->pc.value);
}

/* st_v vr,(sr) */
static void h_st_v_vr_ind_sr(MPE *mpe,UINT32 *args)
{
    scalarRegister *a = (scalarRegister *)args[0];
    MScalar b = ((scalarRegister *)args[1])->value;
    b &= vectorAddressMask;
    mpeStoreScalar(mpe,b,     (*a++).value);
    mpeStoreScalar(mpe,b + 4 ,(*a++).value);
    mpeStoreScalar(mpe,b + 8 ,(*a++).value);
    mpeStoreScalar(mpe,b + 12,(*a++).value);
}

/* st_v vr,dt */
static void h_st_v_vr_dsoff16(MPE *mpe,UINT32 *args)
{
    scalarRegister *a = (scalarRegister *)args[0];
    MScalar b = (MScalar)args[2];
    b += (args[1] == dtRAMSpace ? localRAMBase : localROMBase);
    b &= vectorAddressMask;
    mpeStoreScalar(mpe,b,     (*a++).value);
    mpeStoreScalar(mpe,b + 4 ,(*a++).value);
    mpeStoreScalar(mpe,b + 8 ,(*a++).value);
    mpeStoreScalar(mpe,b + 12,(*a++).value);
}

/* st_sv vr,(sr) */
static void h_st_sv_vr_ind_sr(MPE *mpe,UINT32 *args)
{
    scalarRegister *a = (scalarRegister *)args[0];
    MScalar b = ((scalarRegister *)args[1])->value;
    MScalar value;
    b &= smallVectorAddressMask;
    value = (*a++).value >> 16;
    value |= (*a++).value & 0xFFFF0000;
    mpeStoreScalar(mpe,b,    value);
    value = (*a++).value >> 16;
    value |= (*a++).value & 0xFFFF0000;
    mpeStoreScalar(mpe,b + 4,(*a++).value);
}

/* st_sv vr,dt */
static void h_st_sv_vr_dsoff8(MPE *mpe,UINT32 *args)
{
    scalarRegister *a = (scalarRegister *)args[0];
    MScalar b = (MScalar)args[2];
    MScalar value;
    b += (args[1] == dtRAMSpace ? localRAMBase : localROMBase);
    b &= smallVectorAddressMask;
    value = (*a++).value >> 16;
    value |= (*a++).value & 0xFFFF0000;
    mpeStoreScalar(mpe,b,    value);
    value = (*a++).value >> 16;
    value |= (*a++).value & 0xFFFF0000;
    mpeStoreScalar(mpe,b + 4,(*a++).value);
}

/* st_sv vr,(xy) */
static void h_st_sv_vr_ind_bilinear(MPE *mpe,UINT32 *args)
{
    CallHaltHook(mpe,hhError,mpe->pc.value);
}

/* st_p pr,(xy) */
static void h_st_p_pr_ind_bilinear(MPE *mpe,UINT32 *args)
{
    scalarRegister *r = (scalarRegister *)args[0];
    MScalar ea,pixel,v0,v1,v2,v3;

    /* build the pixel components */
    v0 = ((*r++).value >> 22);
    v1 = ((*r++).value >> 14);
    v2 = ((*r++).value >>  6);
    v3 = ((*r++).value <<  2);

    /* normalize the chrominance values */
    if (TRUE) {
        v1 += 0x00008000;
        v2 += 0x00800000;
    }
    
    /* build the pixel */
    pixel = (v0 & 0x000000ff) | (v1 & 0x0000ff00)
          | (v2 & 0x00ff0000) | (v3 & 0xff000000);

    /* compute the pixel address and store the pixel */
    ComputePixelAddress((bilinearRegister *)args[1],4,&ea);
    mpeStoreScalar(mpe,ea,pixel);
}

/* st_io sr,(io) */
static void h_st_io_sr_ind_io(MPE *mpe,UINT32 *args)
{
    scalarRegister *a = (scalarRegister *)args[0];
    ioRegister *b = (ioRegister *)args[1];
    b->newValue = a->value;
    if (b->writeHandler)
        (*b->writeHandler)(mpe);
}

/* st_io vr,(io) */
static void h_st_io_vr_ind_io(MPE *mpe,UINT32 *args)
{
    scalarRegister *a = (scalarRegister *)args[0];
    ioVectorRegister *b = (ioVectorRegister *)args[1];
    ioRegister **p = b->data;
    (*p++)->newValue = (*a++).value;
    (*p++)->newValue = (*a++).value;
    (*p++)->newValue = (*a++).value;
    (*p++)->newValue = (*a++).value;
    if (b->writeHandler)
        (*b->writeHandler)(mpe);
}

/* st_io #n,(io) */
static void h_st_io_imm_ind_io(MPE *mpe,UINT32 *args)
{
    MScalar a = (MScalar)args[0];
    ioRegister *b = (ioRegister *)args[1];
    b->newValue = a;
    if (b->writeHandler)
        (*b->writeHandler)(mpe);
}

/* mv_s sr,sr */
static void h_mv_s_sr_sr(MPE *mpe,UINT32 *args)
{
    scalarRegister *a = (scalarRegister *)args[0];
    scalarRegister *b = (scalarRegister *)args[1];
    b->newValue = a->value;
}

/* mv_s #n,sr */
static void h_mv_s_imm_sr(MPE *mpe,UINT32 *args)
{
    MScalar a = (MScalar)args[0];
    scalarRegister *b = (scalarRegister *)args[1];
    b->newValue = a;
}

/* mv_s spr,sr */
static void h_mv_s_spr_sr(MPE *mpe,UINT32 *args)
{
    ioRegister *a = (ioRegister *)args[0];
    scalarRegister *b = (scalarRegister *)args[1];
    b->newValue = a->value;
}

/* mv_s #n,sr */
static void h_mv_s_imm32_sr(MPE *mpe,UINT32 *args)
{
    MScalar a = (MScalar)args[0];
    scalarRegister *b = (scalarRegister *)args[1];
    b->newValue = a;
}

/* mv_v vr,vr */
static void h_mv_v_vr_vr(MPE *mpe,UINT32 *args)
{
    scalarRegister *a = (scalarRegister *)args[0];
    scalarRegister *b = (scalarRegister *)args[1];
    (*b++).newValue = (*a++).value;
    (*b++).newValue = (*a++).value;
    (*b++).newValue = (*a++).value;
    (*b++).newValue = (*a++).value;
}

/* push vr */
static void h_push_vr(MPE *mpe,UINT32 *args)
{
    scalarRegister *a = (scalarRegister *)args[0];
    MScalar p = mpe->sp.value - 16;
    mpeStoreScalar(mpe,p,     (*a++).value);
    mpeStoreScalar(mpe,p + 4 ,(*a++).value);
    mpeStoreScalar(mpe,p + 8 ,(*a++).value);
    mpeStoreScalar(mpe,p + 12,(*a++).value);
    mpe->sp.newValue = p;
}

/* push vr,rz */
static void h_push_vr_rz(MPE *mpe,UINT32 *args)
{
    scalarRegister *a = (scalarRegister *)args[0];
    MScalar p = mpe->sp.value - 16;
    mpeStoreScalar(mpe,p,     (*a++).value);
    mpeStoreScalar(mpe,p + 4 ,(*a++).value);
    mpeStoreScalar(mpe,p + 8 ,(*a++).value);
    mpeStoreScalar(mpe,p + 12,mpe->rz.value);
    mpe->sp.newValue = p;
}

/* pop vr */
static void h_pop_vr(MPE *mpe,UINT32 *args)
{
    scalarRegister *a = (scalarRegister *)args[0];
    MScalar p = mpe->sp.value;
    mpeFetchScalar(mpe,p,     &(*a++).newValue);
    mpeFetchScalar(mpe,p + 4 ,&(*a++).newValue);
    mpeFetchScalar(mpe,p + 8 ,&(*a++).newValue);
    mpeFetchScalar(mpe,p + 12,&(*a++).newValue);
    mpe->sp.newValue = p + 16;
}

/* pop vr,rz */
static void h_pop_vr_rz(MPE *mpe,UINT32 *args)
{
    scalarRegister *a = (scalarRegister *)args[0];
    MScalar p = mpe->sp.value;
    mpeFetchScalar(mpe,p,     &(*a++).newValue);
    mpeFetchScalar(mpe,p + 4 ,&(*a++).newValue);
    mpeFetchScalar(mpe,p + 8 ,&(*a++).newValue);
    mpeFetchScalar(mpe,p + 12,&mpe->rz.newValue);
    mpe->sp.newValue = p + 16;
}

/* bra cc,#n */
static void h_bra_cc_imm(MPE *mpe,UINT32 *args)
{
    int cc = (int)args[0];
    MScalar offset = (MScalar)args[1];
    Branch(mpe,cc,btRelative,offset,dsDelay);
}

/* bra #n */
static void h_bra_imm(MPE *mpe,UINT32 *args)
{
    MScalar offset = (MScalar)args[0];
    Branch(mpe,ccT,btRelative,offset,dsDelay);
}

/* jmp cc,#n,nop */
static void h_jmp_cc_imm_stallop(MPE *mpe,UINT32 *args)
{
    int cc = (int)args[0];
    MScalar ea = (MScalar)args[1];
    int stallop = args[2] == soNOP ? dsImpliedNOP : dsDelay;
    Branch(mpe,cc,btAbsolute,ea,stallop);
}

/* jmp cc,(sr),nop */
static void h_jmp_cc_ind_sr_stallop(MPE *mpe,UINT32 *args)
{
    int cc = (int)args[0];
    MScalar ea = ((scalarRegister *)args[1])->value;
    int stallop = args[2] == soNOP ? dsImpliedNOP : dsDelay;
    Branch(mpe,cc,btAbsolute,ea,stallop);
}

/* rts cc,(rz) */
static void h_rts_cc_ind_rz(MPE *mpe,UINT32 *args)
{
    int cc = (int)args[0];
    ioRegister *r = (ioRegister *)args[1];
    MScalar ea = r->value;
    Branch(mpe,cc,btAbsolute,ea,dsDelay);
}

/* Branch - common branch routine */
static void Branch(MPE *mpe,int cc,int type,MScalar target,int stallop)
{
    mpe->branchType = type;
    mpe->branchCondition = cc;
    mpe->branchTarget = target;
    mpe->branchDelaySlot = stallop;
}

/* halt (a fake instruction!)  */
static void h_halt(MPE *mpe,UINT32 *args)
{
    MScalar addr = mpe->pc.value;
    mpe->pc.newValue = addr + bytesPerInstruction;
    LatchNewValues(mpe);
    if (!CallHaltHook(mpe,hhHalt,addr))
        mpe->runState = rsStop;
}

/* opcode to handler table */
OpcodeEntry optable[] = {
{ "nop",                           op_nop,
                                   ifALU,
                                   h_nop                           },
{ "abs_sr",                        op_abs_sr,
                                   ifALU | ifArg1IsSR,
                                   h_abs_sr                        },
{ "add_sr_sr",                     op_add_sr_sr,
                                   ifALU | ifArg1IsSR | ifArg2IsSR,
                                   h_add_sr_sr                     },
{ "add_uimm_sr",                   op_add_uimm_sr,
                                   ifALU | ifArg2IsSR,
                                   h_add_uimm_sr                   },
{ "add_sr_sr_sr",                  op_add_sr_sr_sr,
                                   ifALU | if32Bits | ifArg1IsSR | ifArg2IsSR | ifArg3IsSR,
                                   h_add_sr_sr_sr                  },
{ "add_uimm_sr_sr",                op_add_uimm_sr_sr,
                                   ifALU | if32Bits | ifArg2IsSR | ifArg3IsSR,
                                   h_add_uimm_sr_sr                },
{ "add_uimm10_sr",                 op_add_uimm10_sr,
                                   ifALU | if32Bits | ifArg2IsSR,
                                   h_add_uimm10_sr                 },
{ "add_uimm_rotate_imm_sr",        op_add_uimm_rotate_imm_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg3IsSR,
                                   h_add_uimm_rotate_imm_sr        },
{ "add_sr_uimm_sr",                op_add_sr_uimm_sr,
                                   ifALU | if32Bits | ifArg1IsSR | ifArg3IsSR,
                                   h_add_sr_uimm_sr                },
{ "add_uimm_shift_sr_sr",          op_add_uimm_shift_sr_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg2IsSR | ifArg3IsSR,
                                   h_add_uimm_shift_sr_sr          },
{ "add_sr_shift_imm_sr",           op_add_sr_shift_imm_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg1IsSR | ifArg3IsSR,
                                   h_add_sr_shift_imm_sr           },
{ "add_sr_shift_sr_sr",            op_add_sr_shift_sr_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg1IsSR | ifArg2IsSR | ifArg3IsSR,
                                   h_add_sr_shift_sr_sr            },
{ "add_sv_vr_vr",                  op_add_sv_vr_vr,
                                   ifALU | ifArg1IsVR | ifArg2IsVR,
                                   h_add_sv_vr_vr                  },
{ "add_sv_vr_vr_vr",               op_add_sv_vr_vr_vr,
                                   ifALU | if32Bits | ifArg1IsVR | ifArg2IsVR | ifArg3IsVR,
                                   h_add_sv_vr_vr_vr               },
{ "sub_sr_sr",                     op_sub_sr_sr,
                                   ifALU | ifArg1IsSR | ifArg2IsSR,
                                   h_sub_sr_sr                     },
{ "sub_uimm_sr",                   op_sub_uimm_sr,
                                   ifALU | ifArg2IsSR,
                                   h_sub_uimm_sr                   },
{ "sub_sr_sr_sr",                  op_sub_sr_sr_sr,
                                   ifALU | if32Bits | ifArg1IsSR | ifArg2IsSR | ifArg3IsSR,
                                   h_sub_sr_sr_sr                  },
{ "sub_uimm_sr_sr",                op_sub_uimm_sr_sr,
                                   ifALU | if32Bits | ifArg2IsSR | ifArg3IsSR,
                                   h_sub_uimm_sr_sr                },
{ "sub_uimm10_sr",                 op_sub_uimm10_sr,
                                   ifALU | if32Bits | ifArg2IsSR,
                                   h_sub_uimm10_sr                 },
{ "sub_uimm_rotate_imm_sr",        op_sub_uimm_rotate_imm_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg3IsSR,
                                   h_sub_uimm_rotate_imm_sr        },
{ "sub_sr_uimm_sr",                op_sub_sr_uimm_sr,
                                   ifALU | if32Bits | ifArg1IsSR | ifArg3IsSR,
                                   h_sub_sr_uimm_sr                },
{ "sub_uimm_shift_sr_sr",          op_sub_uimm_shift_sr_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg2IsSR | ifArg3IsSR,
                                   h_sub_uimm_shift_sr_sr          },
{ "sub_sr_shift_imm_sr",           op_sub_sr_shift_imm_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg1IsSR | ifArg3IsSR,
                                   h_sub_sr_shift_imm_sr           },
{ "sub_sr_shift_sr_sr",            op_sub_sr_shift_sr_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg1IsSR | ifArg2IsSR | ifArg3IsSR,
                                   h_sub_sr_shift_sr_sr            },
{ "sub_sv_vr_vr",                  op_sub_sv_vr_vr,
                                   ifALU | ifArg1IsVR | ifArg2IsVR,
                                   h_sub_sv_vr_vr                  },
{ "sub_sv_vr_vr_vr",               op_sub_sv_vr_vr_vr,
                                   ifALU | if32Bits | ifArg1IsVR | ifArg2IsVR | ifArg3IsVR,
                                   h_sub_sv_vr_vr_vr               },
{ "cmp_sr_sr",                     op_cmp_sr_sr,
                                   ifALU | ifArg1IsSR | ifArg2IsSR,
                                   h_cmp_sr_sr                     },
{ "cmp_uimm_sr",                   op_cmp_uimm_sr,
                                   ifALU | ifArg2IsSR,
                                   h_cmp_uimm_sr                   },
{ "cmp_uimm10_sr",                 op_cmp_uimm10_sr,
                                   ifALU | if32Bits | ifArg2IsSR,
                                   h_cmp_uimm10_sr                 },
{ "cmp_uimm_rotate_imm_sr",        op_cmp_uimm_rotate_imm_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg3IsSR,
                                   h_cmp_uimm_rotate_imm_sr        },
{ "cmp_sr_uimm",                   op_cmp_sr_uimm,
                                   ifALU | if32Bits | ifArg1IsSR,
                                   h_cmp_sr_uimm                   },
{ "cmp_uimm_shift_sr_sr",          op_cmp_uimm_shift_sr_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg2IsSR | ifArg3IsSR,
                                   h_cmp_uimm_shift_sr_sr          },
{ "cmp_sr_shift_imm_sr",           op_cmp_sr_shift_imm_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg1IsSR | ifArg3IsSR,
                                   h_cmp_sr_shift_imm_sr           },
{ "cmp_sr_shift_sr_sr",            op_cmp_sr_shift_sr_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg1IsSR | ifArg2IsSR | ifArg3IsSR,
                                   h_cmp_sr_shift_sr_sr            },
{ "and_sr_sr",                     op_and_sr_sr,
                                   ifALU | ifArg1IsSR | ifArg2IsSR,
                                   h_and_sr_sr                     },
{ "and_sr_sr_sr",                  op_and_sr_sr_sr,
                                   ifALU | if32Bits | ifArg1IsSR | ifArg2IsSR | ifArg3IsSR,
                                   h_and_sr_sr_sr                  },
{ "and_imm_sr_sr",                 op_and_imm_sr_sr,
                                   ifALU | if32Bits | ifArg2IsSR | ifArg3IsSR,
                                   h_and_imm_sr_sr                 },
{ "and_imm_rotate_imm_sr",         op_and_imm_rotate_imm_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg3IsSR,
                                   h_and_imm_rotate_imm_sr         },
{ "and_imm_shift_sr_sr",           op_and_imm_shift_sr_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg2IsSR | ifArg3IsSR,
                                   h_and_imm_shift_sr_sr           },
{ "and_sr_shift_imm_sr",           op_and_sr_shift_imm_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg1IsSR | ifArg3IsSR,
                                   h_and_sr_shift_imm_sr           },
{ "and_sr_shift_sr_sr",            op_and_sr_shift_sr_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg1IsSR | ifArg2IsSR | ifArg3IsSR,
                                   h_and_sr_shift_sr_sr            },
{ "and_sr_rotate_sr_sr",           op_and_sr_rotate_sr_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg1IsSR | ifArg2IsSR | ifArg3IsSR,
                                   h_and_sr_rotate_sr_sr           },
{ "or_sr_sr",                      op_or_sr_sr,
                                   ifALU | ifArg1IsSR | ifArg2IsSR,
                                   h_or_sr_sr                      },
{ "or_sr_sr_sr",                   op_or_sr_sr_sr,
                                   ifALU | if32Bits | ifArg1IsSR | ifArg2IsSR | ifArg3IsSR,
                                   h_or_sr_sr_sr                   },
{ "or_imm_sr_sr",                  op_or_imm_sr_sr,
                                   ifALU | if32Bits | ifArg2IsSR | ifArg3IsSR,
                                   h_or_imm_sr_sr                  },
{ "or_imm_rotate_imm_sr",          op_or_imm_rotate_imm_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg3IsSR,
                                   h_or_imm_rotate_imm_sr          },
{ "or_imm_shift_sr_sr",            op_or_imm_shift_sr_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg2IsSR | ifArg3IsSR,
                                   h_or_imm_shift_sr_sr            },
{ "or_sr_shift_imm_sr",            op_or_sr_shift_imm_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg1IsSR | ifArg3IsSR,
                                   h_or_sr_shift_imm_sr            },
{ "or_sr_shift_sr_sr",             op_or_sr_shift_sr_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg1IsSR | ifArg2IsSR | ifArg3IsSR,
                                   h_or_sr_shift_sr_sr             },
{ "or_sr_rotate_sr_sr",            op_or_sr_rotate_sr_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg1IsSR | ifArg2IsSR | ifArg3IsSR,
                                   h_or_sr_rotate_sr_sr            },
{ "eor_sr_sr",                     op_eor_sr_sr,
                                   ifALU | ifArg1IsSR | ifArg2IsSR,
                                   h_eor_sr_sr                     },
{ "eor_sr_sr_sr",                  op_eor_sr_sr_sr,
                                   ifALU | if32Bits | ifArg1IsSR | ifArg2IsSR | ifArg3IsSR,
                                   h_eor_sr_sr_sr                  },
{ "eor_imm_sr_sr",                 op_eor_imm_sr_sr,
                                   ifALU | if32Bits | ifArg2IsSR | ifArg3IsSR,
                                   h_eor_imm_sr_sr                 },
{ "eor_imm_rotate_imm_sr",         op_eor_imm_rotate_imm_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg3IsSR,
                                   h_eor_imm_rotate_imm_sr         },
{ "eor_imm_shift_sr_sr",           op_eor_imm_shift_sr_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg2IsSR | ifArg3IsSR,
                                   h_eor_imm_shift_sr_sr           },
{ "eor_sr_shift_imm_sr",           op_eor_sr_shift_imm_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg1IsSR | ifArg3IsSR,
                                   h_eor_sr_shift_imm_sr           },
{ "eor_sr_shift_sr_sr",            op_eor_sr_shift_sr_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg1IsSR | ifArg2IsSR | ifArg3IsSR,
                                   h_eor_sr_shift_sr_sr            },
{ "eor_sr_rotate_sr_sr",           op_eor_sr_rotate_sr_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg1IsSR | ifArg2IsSR | ifArg3IsSR,
                                   h_eor_sr_rotate_sr_sr           },
{ "neg_sr",                        op_neg_sr,
                                   ifALU | ifArg1IsSR,
                                   h_neg_sr                        },
{ "not_sr",                        op_not_sr,
                                   ifALU | ifArg1IsSR,
                                   h_not_sr                        },
{ "bits_uimm_shift_uimm_sr",       op_bits_uimm_shift_uimm_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg3IsSR,
                                   h_bits_uimm_shift_uimm_sr       },
{ "as_sr_sr_sr",                   op_as_sr_sr_sr,
                                   ifALU | if32Bits | ifArg1IsSR | ifArg2IsSR | ifArg3IsSR,
                                   h_as_sr_sr_sr                   },
{ "asl_uimm_sr",                   op_asl_uimm_sr,
                                   ifALU | ifArg2IsSR,
                                   h_asl_uimm_sr                   },
{ "asl_uimm_sr_sr",                op_asl_uimm_sr_sr,
                                   ifALU | if32Bits | ifArg2IsSR | ifArg3IsSR,
                                   h_asl_uimm_sr_sr                },
{ "asr_uimm_sr",                   op_asr_uimm_sr,
                                   ifALU | ifArg2IsSR,
                                   h_asr_uimm_sr                   },
{ "asr_uimm_sr_sr",                op_asr_uimm_sr_sr,
                                   ifALU | if32Bits | ifArg2IsSR | ifArg3IsSR,
                                   h_asr_uimm_sr_sr                },
{ "ls_sr_sr_sr",                   op_ls_sr_sr_sr,
                                   ifALU | if32Bits | ifArg1IsSR | ifArg2IsSR | ifArg3IsSR,
                                   h_ls_sr_sr_sr                   },
{ "lsl_uimm_sr",                   op_lsl_uimm_sr,
                                   ifALU | ifArg2IsSR,
                                   h_lsl_uimm_sr                   },
{ "lsl_uimm_sr_sr",                op_lsl_uimm_sr_sr,
                                   ifALU | if32Bits | ifArg2IsSR | ifArg3IsSR,
                                   h_lsl_uimm_sr_sr                },
{ "lsr_uimm_sr",                   op_lsr_uimm_sr,
                                   ifALU | ifArg2IsSR,
                                   h_lsr_uimm_sr                   },
{ "lsr_uimm_sr_sr",                op_lsr_uimm_sr_sr,
                                   ifALU | if32Bits | ifArg2IsSR | ifArg3IsSR,
                                   h_lsr_uimm_sr_sr                },
{ "msb_sr_sr",                     op_msb_sr_sr,
                                   ifALU | ifArg1IsSR | ifArg2IsSR,
                                   h_msb_sr_sr                     },
{ "msb_sr_sr_sr",                  op_msb_sr_sr_sr,
                                   ifALU | if32Bits | ifArg1IsSR | ifArg2IsSR | ifArg3IsSR,
                                   h_msb_sr_sr_sr                  },
{ "msb_imm_sr_sr",                 op_msb_imm_sr_sr,
                                   ifALU | if32Bits | ifArg2IsSR | ifArg3IsSR,
                                   h_msb_imm_sr_sr                 },
{ "bclr_uimm_sr",                  op_bclr_uimm_sr,
                                   ifALU | ifArg2IsSR,
                                   h_bclr_uimm_sr                  },
{ "bset_uimm_sr",                  op_bset_uimm_sr,
                                   ifALU | ifArg2IsSR,
                                   h_bset_uimm_sr                  },
{ "btst_uimm_sr",                  op_btst_uimm_sr,
                                   ifALU | ifArg2IsSR,
                                   h_btst_uimm_sr                  },
{ "btst_imm_rotate_imm_sr",        op_btst_imm_rotate_imm_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg3IsSR,
                                   h_btst_imm_rotate_imm_sr        },
{ "btst_imm_shift_sr_sr",          op_btst_imm_shift_sr_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg2IsSR | ifArg3IsSR,
                                   h_btst_imm_shift_sr_sr          },
{ "btst_sr_shift_imm_sr",          op_btst_sr_shift_imm_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg1IsSR | ifArg3IsSR,
                                   h_btst_sr_shift_imm_sr          },
{ "btst_sr_shift_sr_sr",           op_btst_sr_shift_sr_sr,
                                   ifALU | if32Bits | ifALUShifter | ifArg1IsSR | ifArg2IsSR | ifArg3IsSR,
                                   h_btst_sr_shift_sr_sr           },
{ "copy_sr_sr",                    op_copy_sr_sr,
                                   ifALU | ifArg1IsSR | ifArg2IsSR,
                                   h_copy_sr_sr                    },
{ "butt_sr_sr",                    op_butt_sr_sr,
                                   ifALU | ifArg1IsHR | ifArg2IsHR,
                                   h_butt_sr_sr                    },
{ "butt_sr_sr_sr",                 op_butt_sr_sr_sr,
                                   ifALU | if32Bits | ifArg1IsSR | ifArg2IsSR | ifArg3IsHR,
                                   h_butt_sr_sr_sr                 },
{ "dotp_sr_vr_shift_sv_shift_sr",  op_dotp_sr_vr_shift_sv_shift_sr,
                                   ifMultiplier | if32Bits | ifALUShifter | ifArg1IsSR | ifArg2IsVR | ifArg4IsSR,
                                   h_dotp_sr_vr_shift_sv_shift_sr  },
{ "dotp_sr_vr_shift_mulimm_sr",    op_dotp_sr_vr_shift_mulimm_sr,
                                   ifMultiplier | if32Bits | ifALUShifter | ifArg1IsSR | ifArg2IsVR | ifArg4IsSR,
                                   h_dotp_sr_vr_shift_mulimm_sr    },
{ "dotp_vr_vr_shift_sv_shift_sr",  op_dotp_vr_vr_shift_sv_shift_sr,
                                   ifMultiplier | ifALUShifter | ifArg1IsVR | ifArg2IsVR | ifArg4IsSR,
                                   h_dotp_vr_vr_shift_sv_shift_sr  },
{ "dotp_vr_vr_shift_mulimm_sr",    op_dotp_vr_vr_shift_mulimm_sr,
                                   ifMultiplier | ifALUShifter | ifArg1IsVR | ifArg2IsVR | ifArg4IsSR,
                                   h_dotp_vr_vr_shift_mulimm_sr    },
{ "mul_sr_sr",                     op_mul_sr_sr,
                                   ifMultiplier | ifArg1IsSR | ifArg2IsSR,
                                   h_mul_sr_sr                     },
{ "mul_sr_sr_shift_ac_shift_sr",   op_mul_sr_sr_shift_ac_shift_sr,
                                   ifMultiplier | if32Bits | ifALUShifter | ifArg1IsSR | ifArg2IsSR | ifArg4IsSR,
                                   h_mul_sr_sr_shift_ac_shift_sr   },
{ "mul_sr_sr_shift_uimm_sr",       op_mul_sr_sr_shift_uimm_sr,
                                   ifMultiplier | if32Bits | ifALUShifter | ifArg1IsSR | ifArg2IsSR,
                                   h_mul_sr_sr_shift_uimm_sr       },
{ "mul_sr_sr_shift_sr_sr",         op_mul_sr_sr_shift_sr_sr,
                                   ifMultiplier | if32Bits | ifALUShifter | ifArg1IsSR | ifArg2IsSR | ifArg3IsSR,
                                   h_mul_sr_sr_shift_sr_sr         },
{ "mul_uimm_sr_shift_ac_shift_sr", op_mul_uimm_sr_shift_ac_shift_sr,
                                   ifMultiplier | if32Bits | ifALUShifter | ifArg2IsSR | ifArg4IsSR,
                                   h_mul_uimm_sr_shift_ac_shift_sr },
{ "mul_uimm_sr_shift_uimm_sr",     op_mul_uimm_sr_shift_uimm_sr,
                                   ifMultiplier | if32Bits | ifALUShifter | ifArg2IsSR,
                                   h_mul_uimm_sr_shift_uimm_sr     },
{ "mul_uimm_sr_shift_sr_sr",       op_mul_uimm_sr_shift_sr_sr,
                                   ifMultiplier | if32Bits | ifALUShifter | ifArg2IsSR | ifArg3IsSR,
                                   h_mul_uimm_sr_shift_sr_sr       },
{ "mul_p_sr_vr_shift_sv_shift_vr", op_mul_p_sr_vr_shift_sv_shift_vr,
                                   ifMultiplier | if32Bits | ifALUShifter | ifArg1IsSR | ifArg2IsVR | ifArg4IsVR,
                                   h_mul_p_sr_vr_shift_sv_shift_vr },
{ "mul_p_sr_vr_shift_mulimm_vr",   op_mul_p_sr_vr_shift_mulimm_vr,
                                   ifMultiplier | if32Bits | ifALUShifter | ifArg1IsSR | ifArg2IsVR | ifArg4IsVR,
                                   h_mul_p_sr_vr_shift_mulimm_vr   },
{ "mul_p_ruv_vr_shift_sv_shift_vr",op_mul_p_ruv_vr_shift_sv_shift_vr,
                                   ifMultiplier | if32Bits | ifALUShifter | ifArg2IsVR | ifArg4IsVR,
                                   h_mul_p_ruv_vr_shift_sv_shift_vr},
{ "mul_p_ruv_vr_shift_mulimm_vr",  op_mul_p_ruv_vr_shift_mulimm_vr,
                                   ifMultiplier | if32Bits | ifALUShifter | ifArg2IsVR | ifArg4IsVR,
                                   h_mul_p_ruv_vr_shift_mulimm_vr  },
{ "mul_p_vr_vr_shift_sv_shift_vr", op_mul_p_vr_vr_shift_sv_shift_vr,
                                   ifMultiplier | ifALUShifter | ifArg1IsVR | ifArg2IsVR | ifArg4IsVR,
                                   h_mul_p_vr_vr_shift_sv_shift_vr },
{ "mul_p_vr_vr_shift_mulimm_vr",   op_mul_p_vr_vr_shift_mulimm_vr,
                                   ifMultiplier | ifALUShifter | ifArg1IsVR | ifArg2IsVR | ifArg4IsVR,
                                   h_mul_p_vr_vr_shift_mulimm_vr   },
{ "mul_sv_vr_vr",                  op_mul_sv_vr_vr,
                                   ifMultiplier | ifArg1IsVR | ifArg2IsVR,
                                   h_mul_sv_vr_vr                  },
{ "mul_sv_sr_vr_shift_sv_shift_vr",op_mul_sv_sr_vr_shift_sv_shift_vr,
                                   ifMultiplier | if32Bits | ifALUShifter | ifArg1IsSR | ifArg2IsVR | ifArg4IsVR,
                                   h_mul_sv_sr_vr_shift_sv_shift_vr},
{ "mul_sv_sr_vr_shift_mulimm_vr",  op_mul_sv_sr_vr_shift_mulimm_vr,
                                   ifMultiplier | if32Bits | ifALUShifter | ifArg1IsSR | ifArg2IsVR | ifArg4IsVR,
                                   h_mul_sv_sr_vr_shift_mulimm_vr  },
{ "mul_sv_ruv_vr_shift_sv_shift_vr", op_mul_sv_ruv_vr_shift_sv_shift_vr,
                                   ifMultiplier | if32Bits | ifALUShifter | ifArg2IsVR | ifArg4IsVR,
                                   h_mul_sv_ruv_vr_shift_sv_shift_vr },
{ "mul_sv_ruv_vr_shift_mulimm_vr", op_mul_sv_ruv_vr_shift_mulimm_vr,
                                   ifMultiplier | if32Bits | ifALUShifter | ifArg2IsVR | ifArg4IsVR,
                                   h_mul_sv_ruv_vr_shift_mulimm_vr },
{ "mul_sv_vr_vr_shift_sv_shift_vr",op_mul_sv_vr_vr_shift_sv_shift_vr,
                                   ifMultiplier | ifALUShifter | ifArg1IsVR | ifArg2IsVR | ifArg4IsVR,
                                   h_mul_sv_vr_vr_shift_sv_shift_vr},
{ "mul_sv_vr_vr_shift_mulimm_vr",  op_mul_sv_vr_vr_shift_mulimm_vr,
                                   ifMultiplier | ifALUShifter | ifArg1IsVR | ifArg2IsVR | ifArg4IsVR,
                                   h_mul_sv_vr_vr_shift_mulimm_vr  },
{ "addm_sr_sr_sr",                 op_addm_sr_sr_sr,
                                   ifMultiplier | if32Bits | ifArg1IsSR | ifArg2IsSR | ifArg3IsSR,
                                   h_addm_sr_sr_sr                 },
{ "subm_sr_sr_sr",                 op_subm_sr_sr_sr,
                                   ifMultiplier | if32Bits | ifArg1IsSR | ifArg2IsSR | ifArg3IsSR,
                                   h_subm_sr_sr_sr                 },
{ "buttm_sr_sr",                   op_buttm_sr_sr,
                                   ifMultiplier | ifArg1IsHR | ifArg2IsHR,
                                   h_buttm_sr_sr                   },
{ "buttm_sr_sr_sr",                op_buttm_sr_sr_sr,
                                   ifMultiplier | if32Bits | ifArg1IsSR | ifArg2IsSR | ifArg3IsHR,
                                   h_buttm_sr_sr_sr                },
{ "addr_ruimm_rur_rc_rc",          op_addr_ruimm_rur_rc_rc,
                                   ifRegister,
                                   h_addr_ruimm_rur_rc_rc          },
{ "addr_sr_rur_rc_rc",             op_addr_sr_rur_rc_rc,
                                   ifRegister | ifArg1IsSR,
                                   h_addr_sr_rur_rc_rc             },
{ "mv_spr_pc_rz_rc_rc",            op_mv_spr_pc_rz_rc_rc,
                                   ifRegister,
                                   h_mv_spr_pc_rz_rc_rc            },
{ "ld_io_ind_io_sr",               op_ld_io_ind_io_sr,
                                   ifMemory | ifArg2IsSR,
                                   h_ld_io_ind_io_sr               },
{ "ld_io_ind_io_vr",               op_ld_io_ind_io_vr,
                                   ifMemory | if32Bits | ifArg2IsVR,
                                   h_ld_io_ind_io_vr               },
{ "ld_p_ind_bilinear_pr",          op_ld_p_ind_bilinear_pr,
                                   ifMemory | if32Bits | ifArg2IsVR,
                                   h_ld_p_ind_bilinear_pr          },
{ "ld_s_ind_sr_sr",                op_ld_s_ind_sr_sr,
                                   ifMemory | ifArg1IsSR | ifArg2IsSR,
                                   h_ld_s_ind_sr_sr                },
{ "ld_s_dsoff4_sr",                op_ld_s_dsoff4_sr,
                                   ifMemory | if32Bits | ifArg2IsSR,
                                   h_ld_s_dsoff4_sr                },
{ "ld_s_ind_bilinear_sr",          op_ld_s_ind_bilinear_sr,
                                   ifMemory | if32Bits | ifArg2IsSR,
                                   h_ld_s_ind_bilinear_sr          },
{ "ld_sv_ind_sr_vr",               op_ld_sv_ind_sr_vr,
                                   ifMemory | if32Bits | ifArg1IsSR | ifArg2IsVR,
                                   h_ld_sv_ind_sr_vr               },
{ "ld_sv_dsoff8_vr",               op_ld_sv_dsoff8_vr,
                                   ifMemory | if32Bits | ifArg2IsVR,
                                   h_ld_sv_dsoff8_vr               },
{ "ld_sv_ind_bilinear_vr",         op_ld_sv_ind_bilinear_vr,
                                   ifMemory | if32Bits | ifArg2IsVR,
                                   h_ld_sv_ind_bilinear_vr         },
{ "ld_v_ind_sr_vr",                op_ld_v_ind_sr_vr,
                                   ifMemory | if32Bits | ifArg1IsSR | ifArg2IsVR,
                                   h_ld_v_ind_sr_vr                },
{ "ld_v_dsoff16_vr",               op_ld_v_dsoff16_vr,
                                   ifMemory | if32Bits | ifArg2IsVR,
                                   h_ld_v_dsoff16_vr               },
{ "st_io_sr_ind_io",               op_st_io_sr_ind_io,
                                   ifMemory | ifArg1IsSR,
                                   h_st_io_sr_ind_io               },
{ "st_io_imm_ind_io",              op_st_io_imm_ind_io,
                                   ifMemory | if32Bits,
                                   h_st_io_imm_ind_io              },
{ "st_io_vr_ind_io",               op_st_io_vr_ind_io,
                                   ifMemory | if32Bits | ifArg1IsVR,
                                   h_st_io_vr_ind_io               },
{ "st_p_pr_ind_bilinear",          op_st_p_pr_ind_bilinear,
                                   ifMemory | if32Bits | ifArg1IsVR,
                                   h_st_p_pr_ind_bilinear          },
{ "st_s_sr_ind_sr",                op_st_s_sr_ind_sr,
                                   ifMemory | ifArg1IsSR | ifArg2IsSR,
                                   h_st_s_sr_ind_sr                },
{ "st_s_sr_dsoff4",                op_st_s_sr_dsoff4,
                                   ifMemory | if32Bits | ifArg1IsSR,
                                   h_st_s_sr_dsoff4                },
{ "st_s_sr_ind_bilinear",          op_st_s_sr_ind_bilinear,
                                   ifMemory | if32Bits | ifArg1IsSR,
                                   h_st_s_sr_ind_bilinear          },
{ "st_sv_vr_ind_sr",               op_st_sv_vr_ind_sr,
                                   ifMemory | if32Bits | ifArg1IsVR | ifArg2IsSR,
                                   h_st_sv_vr_ind_sr               },
{ "st_sv_vr_dsoff8",               op_st_sv_vr_dsoff8,
                                   ifMemory | if32Bits | ifArg1IsVR,
                                   h_st_sv_vr_dsoff8               },
{ "st_sv_vr_ind_bilinear",         op_st_sv_vr_ind_bilinear,
                                   ifMemory | if32Bits | ifArg1IsVR,
                                   h_st_sv_vr_ind_bilinear         },
{ "st_v_vr_ind_sr",                op_st_v_vr_ind_sr,
                                   ifMemory | if32Bits | ifArg1IsVR | ifArg2IsSR,
                                   h_st_v_vr_ind_sr                },
{ "st_v_vr_dsoff16",               op_st_v_vr_dsoff16,
                                   ifMemory | if32Bits | ifArg1IsVR,
                                   h_st_v_vr_dsoff16               },
{ "mv_s_sr_sr",                    op_mv_s_sr_sr,
                                   ifMemory | ifArg1IsSR | ifArg2IsSR,
                                   h_mv_s_sr_sr                    },
{ "mv_s_imm_sr",                   op_mv_s_imm_sr,
                                   ifMemory | ifArg2IsSR,
                                   h_mv_s_imm_sr                   },
{ "mv_s_imm32_sr",                 op_mv_s_imm32_sr,
                                   ifMemory | if48Bits | ifArg2IsSR,
                                   h_mv_s_imm32_sr                 },
{ "mv_v_vr_vr",                    op_mv_v_vr_vr,
                                   ifMemory | ifArg1IsVR | ifArg2IsVR,
                                   h_mv_v_vr_vr                    },
{ "push_vr",                       op_push_vr,
                                   ifMemory | ifArg1IsVR,
                                   h_push_vr                       },
{ "push_vr_rz",                    op_push_vr_rz,
                                   ifMemory | ifArg1IsVR,
                                   h_push_vr_rz                    },
{ "pop_vr",                        op_pop_vr,
                                   ifMemory | ifArg1IsVR,
                                   h_pop_vr                        },
{ "pop_vr_rz",                     op_pop_vr_rz,
                                   ifMemory | ifArg1IsVR,
                                   h_pop_vr_rz                     },
{ "bra_cc_imm",                    op_bra_cc_imm,
                                   ifExecute,
                                   h_bra_cc_imm                    },
{ "bra_imm",                       op_bra_imm,
                                   ifExecute,
                                   h_bra_imm                       },
{ "jmp_cc_imm_stallop",            op_jmp_cc_imm_stallop,
                                   ifExecute,
                                   h_jmp_cc_imm_stallop            },
{ "jmp_cc_ind_sr_stallop",         op_jmp_cc_ind_sr_stallop,
                                   ifExecute | ifArg2IsSR,
                                   h_jmp_cc_ind_sr_stallop         },
{ "rts_cc_ind_rz",                 op_rts_cc_ind_rz,
                                   ifExecute,
                                   h_rts_cc_ind_rz                 },
{ "halt",                          op_halt,
                                   ifExecute,
                                   h_halt                          },
{NULL,0,0,NULL}
};

