/* mptest.c - test routine for MPARSE.DLL */

#include <stdio.h>
#include "mpdll.h"

/* prototypes */
void Usage(void);
void Fatal(char *fmt,...);

/* main - the main routine */
void main(int argc,char *argv[])
{
    mpErrorInfo errInfo;
    char *ifile,*ofile;
    int src,dst,cnt;
    int flags = 0;
    char *p;
    
    /* process the arguments */
    for (src = dst = cnt = 1; src < argc; ++src) {

	/* handle options */
	if (argv[src][0] == '-') {
	    for (p = &argv[src][1]; *p != '\0'; )
	    	switch (*p++) {
                case 't':
                    flags |= mpfOutputTables;
                    break;
                case 'b':
                    flags |= mpfOutputBinary;
                    break;
		default:
	    	    Usage();
		}
	}

	/* handle a filename */
	else {
	    argv[dst++] = argv[src];
	    ++cnt;
	}
    }
    argc = cnt;

    /* make sure there is at least one argument */
    if (argc < 2)
	Usage();

    /* get the arguments */
    ifile = argv[1];
    ofile = argc > 2 ? argv[2] : NULL;

    /* assemble the file */
    switch (mpAssemble(ifile,ofile,flags,&errInfo)) {
    case mpsInputFile:
	Fatal("can't open input file %s",ifile);
	break;
    case mpsOutputFile:
	Fatal("can't create output file %s",ofile);
	break;
    case mpsError:
        fprintf(stderr,"Error: %s\n",errInfo.text);
        fprintf(stderr,"File: %s, Line: %ld, Offset: %ld\n",errInfo.file,errInfo.line,errInfo.offset);
	exit(1);
	break;
    }
    exit(0);
}

/* Usage - display a usage message and exit */
void Usage(void)
{
    fprintf(stderr,"usage: mparse <infile> [<outfile>]\n");
    exit(1);
}

/* Fatal - display a fatal error message and exit */
void Fatal(char *fmt,...)
{
    va_list ap;
    va_start(ap,fmt);
    fprintf(stderr,"error: ");
    vfprintf(stderr,fmt,ap);
    putc('\n',stderr);
    va_end(ap);
    exit(1);
}

